# Copyright 2026 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendDict,
    PyLegendUnion,
    PyLegendList,
    PyLegendSequence,
    PyLegendTuple,
    PyLegendOptional,
    PyLegendCallable,
    TYPE_CHECKING,
)
from pylegend.core.language.pandas_api.pandas_api_custom_expressions import (
    PandasApiPartialFrame,
    PandasApiSortDirection,
    PandasApiSortInfo,
    PandasApiWindow,
    PandasApiWindowReference
)
from pylegend.core.language.shared.primitives.primitive import PyLegendPrimitive
from pylegend.core.sql.metamodel import (
    Expression,
    QuerySpecification,
    SelectItem,
    SingleColumn,
    QualifiedNameReference,
    QualifiedName,
)
from pylegend.core.sql.metamodel_extension import WindowExpression
from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import (
    PandasApiAppliedFunction,
)
from pylegend.core.tds.pandas_api.frames.pandas_api_base_tds_frame import (
    PandasApiBaseTdsFrame,
)
from pylegend.core.language.pandas_api.pandas_api_tds_row import PandasApiTdsRow
from pylegend.core.tds.tds_frame import FrameToSqlConfig, FrameToPureConfig
from pylegend.core.tds.tds_column import TdsColumn, PrimitiveTdsColumn
from pylegend.core.tds.sql_query_helpers import create_sub_query
from pylegend.core.language.shared.helpers import generate_pure_lambda, escape_column_name

if TYPE_CHECKING:
    from pylegend.core.tds.pandas_api.frames.pandas_api_groupby_tds_frame import PandasApiGroupbyTdsFrame


class RankFunction(PandasApiAppliedFunction):
    __base_frame: PyLegendUnion[PandasApiBaseTdsFrame, "PandasApiGroupbyTdsFrame"]
    __axis: PyLegendUnion[str, int]
    __method: str
    __numeric_only: bool
    __na_option: str
    __ascending: bool
    __pct: bool

    __column_expression_and_window_tuples: PyLegendList[
        PyLegendTuple[
            PyLegendTuple[str, PyLegendPrimitive],
            PandasApiWindow
        ]
    ]

    @classmethod
    def name(cls) -> str:
        return "rank"  # pragma: no cover

    def __init__(
            self,
            base_frame: PyLegendUnion[PandasApiBaseTdsFrame, "PandasApiGroupbyTdsFrame"],
            axis: PyLegendUnion[str, int],
            method: str,
            numeric_only: bool,
            na_option: str,
            ascending: bool,
            pct: bool
    ) -> None:
        self.__base_frame = base_frame
        self.__axis = axis
        self.__method = method
        self.__numeric_only = numeric_only
        self.__na_option = na_option
        self.__ascending = ascending
        self.__pct = pct

    def to_sql(self, config: FrameToSqlConfig) -> QuerySpecification:
        temp_column_name_suffix = "__INTERNAL_PYLEGEND_COLUMN__"

        base_query = self.base_frame().to_sql_query_object(config)
        db_extension = config.sql_to_string_generator().get_db_extension()

        new_query: QuerySpecification = create_sub_query(base_query, config, "root")
        new_select_items: list[SelectItem] = []

        for c, window in self.__column_expression_and_window_tuples:
            col_sql_expr: Expression = c[1].to_sql_expression({"r": new_query}, config)

            window_expr = WindowExpression(
                nested=col_sql_expr,
                window=window.to_sql_node(new_query, config),
            )
            new_select_items.append(
                SingleColumn(alias=db_extension.quote_identifier(c[0] + temp_column_name_suffix), expression=window_expr)
            )

        new_query.select.selectItems = new_select_items

        new_query = create_sub_query(new_query, config, "root")

        final_select_items: list[SelectItem] = []
        for col in self.calculate_columns():
            col_name = col.get_name()
            col_expr = QualifiedNameReference(QualifiedName([
                db_extension.quote_identifier("root"), db_extension.quote_identifier(col_name + temp_column_name_suffix)
            ]))
            final_select_items.append(
                SingleColumn(
                    alias=db_extension.quote_identifier(col_name),
                    expression=col_expr
                )
            )

        new_query.select.selectItems = final_select_items

        return new_query

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        self._assert_single_column_in_base_frame()

        c, window = self.__column_expression_and_window_tuples[0]
        col_sql_expr: Expression = c[1].to_sql_expression(frame_name_to_base_query_map, config)
        window_expr = WindowExpression(
            nested=col_sql_expr,
            window=window.to_sql_node(frame_name_to_base_query_map['c'], config),
        )

        return window_expr

    @staticmethod
    def _render_single_column_expression(
            c: PyLegendUnion[PyLegendTuple[str, PyLegendPrimitive]], col_name: str, config: FrameToPureConfig
    ) -> str:
        escaped_col_name: str = escape_column_name(col_name)
        expr_str: str = c[1].to_pure_expression(config)
        return f"{escaped_col_name}:{generate_pure_lambda('p,w,r', expr_str)}"

    def to_pure(self, config: FrameToPureConfig) -> str:
        temp_column_name_suffix: str = "__INTERNAL_PYLEGEND_COLUMN__"

        extend_strs: PyLegendList[str] = []
        for c, window in self.__column_expression_and_window_tuples:
            window_expression = window.to_pure_expression(config)
            col_name = c[0] + temp_column_name_suffix
            extend_strs.append(
                f"->extend({window_expression}, ~{self._render_single_column_expression(c, col_name, config)})"
            )
        extend_str = f"{config.separator(1)}".join(extend_strs)

        project_cols = [
            f"{escape_column_name(c[0])}:p|$p.{escape_column_name(c[0] + temp_column_name_suffix)}"
            for c, _ in self.__column_expression_and_window_tuples
        ]
        joined_project_cols = ("," + config.separator(2)).join(project_cols)
        project_str = (
                f"->project(~[{config.separator(2)}"
                f"{joined_project_cols}"
                f"{config.separator(1)}])"
        )

        return (
                f"{self.base_frame().to_pure(config)}{config.separator(1)}"
                f"{extend_str}{config.separator(1)}"
                f"{project_str}"
        )

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        temp_column_name_suffix = "__INTERNAL_PYLEGEND_COLUMN__"
        self._assert_single_column_in_base_frame()
        c, window = self.__column_expression_and_window_tuples[0]
        return f"$c.{c[0] + temp_column_name_suffix}"

    def base_frame(self) -> PandasApiBaseTdsFrame:
        from pylegend.core.tds.pandas_api.frames.pandas_api_groupby_tds_frame import PandasApiGroupbyTdsFrame
        if isinstance(self.__base_frame, PandasApiGroupbyTdsFrame):
            return self.__base_frame.base_frame()
        return self.__base_frame

    def tds_frame_parameters(self) -> PyLegendList["PandasApiBaseTdsFrame"]:
        return []

    def calculate_columns(self) -> PyLegendSequence["TdsColumn"]:
        from pylegend.core.tds.pandas_api.frames.pandas_api_groupby_tds_frame import PandasApiGroupbyTdsFrame

        def __validate_and_convert_column(col: TdsColumn) -> PyLegendOptional["TdsColumn"]:
            valid_column_types_for_numeric_only = ["Integer", "Float", "Number"]
            if self.__numeric_only and col.get_type() not in valid_column_types_for_numeric_only:
                return None
            if self.__pct:
                new_col = PrimitiveTdsColumn.float_column(col.get_name())
            else:
                new_col = PrimitiveTdsColumn.integer_column(col.get_name())
            return new_col

        new_columns: PyLegendList["TdsColumn"] = []
        if isinstance(self.__base_frame, PandasApiGroupbyTdsFrame):
            grouping_column_names = set([col.get_name() for col in self.__base_frame.get_grouping_columns()])
            selected_columns: PyLegendOptional[PyLegendList[TdsColumn]] = self.__base_frame.get_selected_columns()
            if selected_columns is None:
                for col in self.base_frame().columns():
                    if col.get_name() in grouping_column_names:
                        continue
                    validated_col = __validate_and_convert_column(col)
                    if validated_col is not None:
                        new_columns.append(validated_col)
            else:
                for col in selected_columns:
                    validated_col = __validate_and_convert_column(col)
                    if validated_col is not None:
                        new_columns.append(validated_col)
        else:
            for col in self.base_frame().columns():
                validated_col = __validate_and_convert_column(col)
                if validated_col is not None:
                    new_columns.append(validated_col)

        return new_columns

    def validate(self) -> bool:
        if self.__axis not in [0, "index"]:
            raise NotImplementedError(
                f"The 'axis' parameter of the rank function must be 0 or 'index', but got: axis={self.__axis!r}"
            )

        valid_methods: set[str] = {'min', 'first', 'dense'}
        if self.__method not in valid_methods:
            raise NotImplementedError(
                f"The 'method' parameter of the rank function must be one of {sorted(list(valid_methods))!r},"
                f" but got: method={self.__method!r}"
            )
        elif self.__pct is True and self.__method != 'min':
            raise NotImplementedError(
                "The 'pct=True' parameter of the rank function is only supported with method='min',"
                f" but got: method={self.__method!r}."
            )

        valid_na_options = {'bottom'}
        if self.__na_option not in valid_na_options:
            raise NotImplementedError(
                f"The 'na_option' parameter of the rank function must be one of {sorted(list(valid_na_options))!r},"
                f" but got: na_option={self.__na_option!r}"
            )

        self.__column_expression_and_window_tuples = self.construct_column_expression_and_window_tuples()

        return True

    def construct_column_expression_and_window_tuples(self) -> PyLegendList[
        PyLegendTuple[
            PyLegendTuple[str, PyLegendPrimitive],
            PandasApiWindow
        ]
    ]:
        from pylegend.core.tds.pandas_api.frames.pandas_api_groupby_tds_frame import PandasApiGroupbyTdsFrame
        column_names: list[str] = [col.get_name() for col in self.calculate_columns()]

        lambda_func: PyLegendCallable[
            [PandasApiPartialFrame, PandasApiWindowReference, PandasApiTdsRow],
            PyLegendPrimitive
        ]

        if self.__pct:
            def lambda_func(
                p: PandasApiPartialFrame,
                w: PandasApiWindowReference,
                r: PandasApiTdsRow,
            ) -> PyLegendPrimitive:
                return p.percent_rank(w, r)

        elif self.__method == 'min':
            def lambda_func(
                p: PandasApiPartialFrame,
                w: PandasApiWindowReference,
                r: PandasApiTdsRow,
            ) -> PyLegendPrimitive:
                return p.rank(w, r)

        elif self.__method == 'first':
            def lambda_func(
                p: PandasApiPartialFrame,
                w: PandasApiWindowReference,
                r: PandasApiTdsRow,
            ) -> PyLegendPrimitive:
                return p.row_number(r)

        elif self.__method == 'dense':
            def lambda_func(
                p: PandasApiPartialFrame,
                w: PandasApiWindowReference,
                r: PandasApiTdsRow,
            ) -> PyLegendPrimitive:
                return p.dense_rank(w, r)

        else:
            raise ValueError(
                f"Encountered unsupported method parameter (method={self.__method!r}) in rank function")  # pragma: no cover

        extend_columns = [(column_name, lambda_func) for column_name in column_names]

        partial_frame = PandasApiPartialFrame(base_frame=self.base_frame(), var_name="p")

        column_expression_and_window_tuples: PyLegendList[
            PyLegendTuple[
                PyLegendTuple[str, PyLegendPrimitive],
                PandasApiWindow
            ]
        ] = []

        for extend_column in extend_columns:
            current_column_name: str = extend_column[0]

            partition_by: PyLegendOptional[PyLegendList[str]] = None
            if isinstance(self.__base_frame, PandasApiGroupbyTdsFrame):
                partition_by = [col.get_name() for col in self.__base_frame.get_grouping_columns()]

            tds_row = PandasApiTdsRow.from_tds_frame("r", self.base_frame())
            sort_direction: PandasApiSortDirection
            if self.__ascending:
                sort_direction = PandasApiSortDirection.ASC
            else:
                sort_direction = PandasApiSortDirection.DESC
            order_by = PandasApiSortInfo(current_column_name, sort_direction)

            window = PandasApiWindow(partition_by, [order_by], frame=None)
            window_ref = PandasApiWindowReference(window=window, var_name="w")

            result = extend_column[1](partial_frame, window_ref, tds_row)

            column_expression: PyLegendTuple[str, PyLegendPrimitive] = (current_column_name, result)
            column_expression_and_window_tuples.append((column_expression, window))

        return column_expression_and_window_tuples

    def _assert_single_column_in_base_frame(self) -> None:
        from pylegend.core.tds.pandas_api.frames.pandas_api_groupby_tds_frame import PandasApiGroupbyTdsFrame

        if isinstance(self.__base_frame, PandasApiGroupbyTdsFrame):
            selected_columns = self.__base_frame.get_selected_columns()
            assert selected_columns is not None, "To get an SQL or a pure expression, exactly one column must be selected."
            base_frame_columns = selected_columns
        else:
            base_frame_columns = list(self.__base_frame.columns())

        assert len(base_frame_columns) == 1, (
            "To get an SQL or a pure expression, the base frame must have exactly one column, but got "
            f"{len(base_frame_columns)} columns: {[str(col) for col in base_frame_columns]}"
        )



# Copyright 2026 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING
from pylegend._typing import (
    PyLegendUnion,
    PyLegendTuple,
    PyLegendSequence,
)

if TYPE_CHECKING:
    from pylegend.core.tds.pandas_api.frames.pandas_api_base_tds_frame import PandasApiBaseTdsFrame
    from pylegend.core.tds.pandas_api.frames.pandas_api_tds_frame import PandasApiTdsFrame


__all__: PyLegendSequence[str] = [
    "PandasApiIlocIndexer"
]


class PandasApiIlocIndexer:
    _frame: "PandasApiBaseTdsFrame"

    def __init__(self, frame: "PandasApiBaseTdsFrame") -> None:
        self._frame = frame

    def __getitem__(  # type: ignore
            self,
            key: PyLegendUnion[int, slice, PyLegendTuple[PyLegendUnion[int, slice], ...]]
    ) -> "PandasApiTdsFrame":
        if isinstance(key, tuple):
            if len(key) > 2:
                raise IndexError("Too many indexers")
            elif len(key) == 1:
                rows, cols = key[0], slice(None, None, None)
            else:
                rows, cols = key  # type: ignore
        else:
            rows, cols = key, slice(None, None, None)

        # Row selection
        row_frame = self._handle_row_selection(rows)

        # Column selection
        return self._handle_column_selection(row_frame, cols)

    def _handle_row_selection(self, rows: PyLegendUnion[int, slice]) -> "PandasApiTdsFrame":  # type: ignore
        if isinstance(rows, slice):
            if rows.step is not None and rows.step != 1:
                raise NotImplementedError("iloc with slice step other than 1 is not supported yet in Pandas Api")

            start = rows.start
            stop = rows.stop
            after = stop - 1 if stop is not None else None
            return self._frame.truncate(before=start, after=after)

        elif isinstance(rows, int):
            return self._frame.truncate(before=rows, after=rows)

        else:
            raise NotImplementedError(
                f"iloc supports integer, slice, or tuple of these, but got indexer of type: {type(rows)}"
            )

    def _handle_column_selection(  # type: ignore
            self,
            frame: "PandasApiTdsFrame",
            cols: PyLegendUnion[int, slice]
    ) -> "PandasApiTdsFrame":
        if isinstance(cols, slice):
            if cols.step is not None and cols.step != 1:
                raise NotImplementedError("iloc with slice step other than 1 is not supported yet in Pandas Api")

            all_columns = [c.get_name() for c in frame.columns()]
            selected_columns = all_columns[cols]
            return frame.filter(items=selected_columns)

        elif isinstance(cols, int):
            all_columns = [c.get_name() for c in frame.columns()]
            if not -len(all_columns) <= cols < len(all_columns):
                raise IndexError("single positional indexer is out-of-bounds")
            selected_column = all_columns[cols]
            return frame.filter(items=[selected_column])

        else:
            raise NotImplementedError(
                f"iloc supports integer, slice, or tuple of these, but got indexer of type: {type(cols)}"
            )



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from random import Random
import socket


def generate_dynamic_port() -> int:
    min_port = 50352
    retry_count = 0
    random = Random()
    while retry_count < 10:
        retry_count += 1
        next_port = random.randint(min_port, min_port + 10000)
        if _is_port_not_in_use(next_port):
            return next_port

    raise RuntimeError("Unable to obtain a free port")  # pragma: no cover


def _is_port_not_in_use(port: int) -> bool:
    with socket.socket() as s:
        try:
            s.bind(("127.0.0.1", port))
            return True
        except socket.error:  # pragma: no cover
            return False  # pragma: no cover



# Copyright 2026 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendSequence,
    PyLegendDict,
    PyLegendUnion,
    TYPE_CHECKING,
)
from decimal import Decimal as PythonDecimal
from pylegend.core.language.shared.primitives.number import PyLegendNumber
from pylegend.core.language.shared.expression import PyLegendExpressionDecimalReturn
from pylegend.core.language.shared.literal_expressions import PyLegendDecimalLiteralExpression
from pylegend.core.sql.metamodel import (
    Expression,
    QuerySpecification
)
from pylegend.core.tds.pandas_api.frames.helpers.series_helper import grammar_method
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.core.language.shared.operations.decimal_operation_expressions import (
    PyLegendDecimalAbsoluteExpression,
    PyLegendDecimalAddExpression,
    PyLegendDecimalNegativeExpression,
    PyLegendDecimalSubtractExpression,
    PyLegendDecimalMultiplyExpression,
)
if TYPE_CHECKING:
    from pylegend.core.language.shared.primitives import PyLegendInteger, PyLegendFloat

__all__: PyLegendSequence[str] = [
    "PyLegendDecimal"
]


class PyLegendDecimal(PyLegendNumber):
    __value_copy: PyLegendExpressionDecimalReturn

    def __init__(
            self,
            value: PyLegendExpressionDecimalReturn
    ) -> None:
        self.__value_copy = value
        super().__init__(value)

    @grammar_method
    def __add__(
            self,
            other: PyLegendUnion[
                int, float, PythonDecimal, "PyLegendInteger", "PyLegendFloat", "PyLegendDecimal", "PyLegendNumber"
            ]
    ) -> "PyLegendUnion[PyLegendNumber, PyLegendDecimal]":
        PyLegendNumber.validate_param_to_be_number(other, "Decimal plus (+) parameter")
        if isinstance(other, (PythonDecimal, PyLegendDecimal)):
            other_op = PyLegendDecimal.__convert_to_decimal_expr(other)
            return PyLegendDecimal(PyLegendDecimalAddExpression(self.__value_copy, other_op))
        else:
            return super().__add__(other)

    @grammar_method
    def __radd__(
            self,
            other: PyLegendUnion[
                int, float, PythonDecimal, "PyLegendInteger", "PyLegendFloat", "PyLegendDecimal", "PyLegendNumber"
            ]
    ) -> "PyLegendUnion[PyLegendNumber, PyLegendDecimal]":
        PyLegendNumber.validate_param_to_be_number(other, "Decimal plus (+) parameter")
        if isinstance(other, (PythonDecimal, PyLegendDecimal)):
            other_op = PyLegendDecimal.__convert_to_decimal_expr(other)
            return PyLegendDecimal(PyLegendDecimalAddExpression(other_op, self.__value_copy))
        else:
            return super().__radd__(other)

    @grammar_method
    def __sub__(
            self,
            other: PyLegendUnion[
                int, float, PythonDecimal, "PyLegendInteger", "PyLegendFloat", "PyLegendDecimal", "PyLegendNumber"
            ]
    ) -> "PyLegendUnion[PyLegendNumber, PyLegendDecimal]":
        PyLegendNumber.validate_param_to_be_number(other, "Decimal minus (-) parameter")
        if isinstance(other, (PythonDecimal, PyLegendDecimal)):
            other_op = PyLegendDecimal.__convert_to_decimal_expr(other)
            return PyLegendDecimal(PyLegendDecimalSubtractExpression(self.__value_copy, other_op))
        else:
            return super().__sub__(other)

    @grammar_method
    def __rsub__(
            self,
            other: PyLegendUnion[
                int, float, PythonDecimal, "PyLegendInteger", "PyLegendFloat", "PyLegendDecimal", "PyLegendNumber"
            ]
    ) -> "PyLegendUnion[PyLegendNumber, PyLegendDecimal]":
        PyLegendNumber.validate_param_to_be_number(other, "Decimal minus (-) parameter")
        if isinstance(other, (PythonDecimal, PyLegendDecimal)):
            other_op = PyLegendDecimal.__convert_to_decimal_expr(other)
            return PyLegendDecimal(PyLegendDecimalSubtractExpression(other_op, self.__value_copy))
        else:
            return super().__rsub__(other)

    @grammar_method
    def __mul__(
            self,
            other: PyLegendUnion[
                int, float, PythonDecimal, "PyLegendInteger", "PyLegendFloat", "PyLegendDecimal", "PyLegendNumber"
            ]
    ) -> "PyLegendUnion[PyLegendNumber, PyLegendDecimal]":
        PyLegendNumber.validate_param_to_be_number(other, "Decimal multiply (*) parameter")
        if isinstance(other, (PythonDecimal, PyLegendDecimal)):
            other_op = PyLegendDecimal.__convert_to_decimal_expr(other)
            return PyLegendDecimal(PyLegendDecimalMultiplyExpression(self.__value_copy, other_op))
        else:
            return super().__mul__(other)

    @grammar_method
    def __rmul__(
            self,
            other: PyLegendUnion[
                int, float, PythonDecimal, "PyLegendInteger", "PyLegendFloat", "PyLegendDecimal", "PyLegendNumber"
            ]
    ) -> "PyLegendUnion[PyLegendNumber, PyLegendDecimal]":
        PyLegendNumber.validate_param_to_be_number(other, "Decimal multiply (*) parameter")
        if isinstance(other, (PythonDecimal, PyLegendDecimal)):
            other_op = PyLegendDecimal.__convert_to_decimal_expr(other)
            return PyLegendDecimal(PyLegendDecimalMultiplyExpression(other_op, self.__value_copy))
        else:
            return super().__rmul__(other)

    @grammar_method
    def __abs__(self) -> "PyLegendDecimal":
        return PyLegendDecimal(PyLegendDecimalAbsoluteExpression(self.__value_copy))

    @grammar_method
    def __neg__(self) -> "PyLegendDecimal":
        return PyLegendDecimal(PyLegendDecimalNegativeExpression(self.__value_copy))

    @grammar_method
    def __pos__(self) -> "PyLegendDecimal":
        return self

    @staticmethod
    def __convert_to_decimal_expr(
            val: PyLegendUnion[PythonDecimal, "PyLegendDecimal"]
    ) -> PyLegendExpressionDecimalReturn:
        if isinstance(val, (PythonDecimal)):
            return PyLegendDecimalLiteralExpression(val)
        return val.__value_copy

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return super().to_sql_expression(frame_name_to_base_query_map, config)

    def value(self) -> PyLegendExpressionDecimalReturn:
        return self.__value_copy

    @staticmethod
    def __validate__param_to_be_decimal(params, desc):  # type: ignore
        PyLegendNumber.validate_param_to_be_number(params, desc)



# Copyright 2026 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import atexit
import json
import os
import tempfile
import time
import logging
from pathlib import Path
from http.server import BaseHTTPRequestHandler, HTTPServer
from threading import Thread

import requests
from testcontainers.core.container import DockerContainer  # type: ignore

from pylegend._typing import (
    PyLegendSequence,
    PyLegendOptional,
)
from pylegend.core.request.legend_client import LegendClient
from pylegend.utils.dynamic_port_generator import generate_dynamic_port
from pylegend.core.project_cooridnates import VersionedProjectCoordinates


__all__: PyLegendSequence[str] = [
    "get_local_legend_env",
    "NORTHWIND_PROJECT_COORDINATES",
]

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)
if not LOGGER.handlers:
    _ch = logging.StreamHandler()
    _ch.setFormatter(logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))
    LOGGER.addHandler(_ch)

_SAMPLES_DIR = Path(__file__).resolve().parent
_METADATA_DIR = _SAMPLES_DIR / "resources" / "metadata"

_DEFAULT_IMAGE = "eclipse-temurin:11-jdk"

_LEGEND_ENGINE_VERSION = "4.121.0"
_LEGEND_JAR_URL = (
    f"https://repo1.maven.org/maven2/org/finos/legend/engine/legend-engine-server-http-server/"
    f"{_LEGEND_ENGINE_VERSION}/legend-engine-server-http-server-{_LEGEND_ENGINE_VERSION}-shaded.jar"
)


class LocalLegendEnv:
    def __init__(
        self,
        image: str = _DEFAULT_IMAGE,
        metadata_port: PyLegendOptional[int] = None,
        max_wait_seconds: int = 120,
    ) -> None:
        self._image = image
        self._metadata_port = metadata_port or generate_dynamic_port()
        self._max_wait_seconds = max_wait_seconds

        self._engine_container: PyLegendOptional[DockerContainer] = None
        self._metadata_server: PyLegendOptional[HTTPServer] = None
        self._config_path: PyLegendOptional[str] = None
        self._engine_port: PyLegendOptional[int] = None
        self._testdb_port: PyLegendOptional[int] = None
        self._legend_client: PyLegendOptional[LegendClient] = None
        self._server_jar_path: PyLegendOptional[Path] = None

    @property
    def engine_port(self) -> int:
        if self._engine_port is None:
            raise RuntimeError("Server not started")
        return self._engine_port

    @property
    def legend_client(self) -> LegendClient:
        if self._legend_client is None:
            raise RuntimeError("Server not started")
        return self._legend_client

    def start(self) -> "LocalLegendEnv":
        self._validate_resources()
        self._server_jar_path = _get_server_jar_path()
        self._start_metadata_server()
        self._start_engine_container()
        self._wait_for_engine()
        self._legend_client = LegendClient(
            "127.0.0.1", self._engine_port, secure_http=False  # type: ignore[arg-type]
        )
        return self

    def stop(self) -> None:
        if self._engine_container:
            self._engine_container.stop()
        if self._metadata_server:
            self._metadata_server.shutdown()
        if self._config_path:
            try:
                os.unlink(self._config_path)
                os.rmdir(os.path.dirname(self._config_path))
            except OSError:
                pass
        self._engine_container = self._metadata_server = self._config_path = None
        self._engine_port = self._testdb_port = self._legend_client = None

    def _validate_resources(self) -> None:
        meta = _METADATA_DIR
        if not meta.exists():
            raise FileNotFoundError(
                f"Metadata directory not found at {meta}. "
                "Make sure you are running from a pylegend repository checkout."
            )

    def _start_metadata_server(self) -> None:
        handler_class = type(
            "_Handler",
            (_MetadataServerHandler,),
            {"metadata_dir": str(_METADATA_DIR.resolve())},
        )
        self._metadata_server = HTTPServer(
            ("127.0.0.1", self._metadata_port), handler_class
        )
        t = Thread(target=self._metadata_server.serve_forever, daemon=True)
        t.start()
        LOGGER.info("Metadata server listening on port %d", self._metadata_port)

    def _start_engine_container(self) -> None:
        if self._server_jar_path is None:
            raise RuntimeError("Server JAR path is not set")

        self._engine_port = generate_dynamic_port()
        self._testdb_port = generate_dynamic_port()
        self._config_path = _write_server_config(self._engine_port, "127.0.0.1", self._metadata_port, self._testdb_port)

        self._engine_container = (
            DockerContainer(self._image)
            .with_volume_mapping(str(self._server_jar_path.resolve()), "/legend/server.jar", "ro")
            .with_volume_mapping(self._config_path, "/legend/config.json", "ro")
            .with_env("JAVA_TOOL_OPTIONS", "-Duser.timezone=UTC -Dfile.encoding=UTF-8")
            .with_command("sh -c 'java -jar /legend/server.jar server /legend/config.json'")
        )

        self._engine_container.with_kwargs(network_mode="host")

        LOGGER.info("Starting Legend engine container (%s) …", self._image)
        self._engine_container.start()

        LOGGER.info("Engine container started; responding on host network port = %d", self._engine_port)

    def _wait_for_engine(self) -> None:
        url = f"http://127.0.0.1:{self._engine_port}/api/server/v1/info"
        deadline = time.monotonic() + self._max_wait_seconds
        while time.monotonic() < deadline:
            try:
                requests.get(url, timeout=5).raise_for_status()
                return LOGGER.info("Legend engine is healthy.")
            except Exception:
                time.sleep(3)
        raise RuntimeError(f"Engine not healthy within {self._max_wait_seconds}s at {url}")


class _MetadataServerHandler(BaseHTTPRequestHandler):
    metadata_dir: str = ""

    def do_GET(self) -> None:
        if (
            "/depot/api/projects/org.finos.legend.pylegend/pylegend-northwind-models/versions/"
            "0.0.1-SNAPSHOT/pureModelContextData?convertToNewProtocol=false&clientVersion=v1_33_0"
        ) != self.path:
            return self.send_error(404, f"Unhandled metadata path: {self.path}")

        file_path = os.path.join(self.metadata_dir, "org.finos.legend.pylegend_pylegend-northwind-models_0.0.1-SNAPSHOT.json")
        try:
            with open(file_path, "rb") as f:
                content = f.read()
            self.send_response(200)
            self.send_header("Content-Type", "application/json; charset=utf-8")
            self.send_header("Content-Length", str(len(content)))
            self.end_headers()
            self.wfile.write(content)
        except OSError:
            self.send_error(404, f"Metadata file not found: {file_path}")

    def log_message(self, format: str, *args: object) -> None: pass


def _write_server_config(engine_port: int, metadata_host: str, metadata_port: int, testdb_port: int) -> str:
    with open(_SAMPLES_DIR / "resources" / "config" / "engine_config.json", "r", encoding="utf-8") as f:
        config = json.load(f)

    config["server"]["connector"]["port"] = engine_port
    config["server"]["connector"]["bindHost"] = "127.0.0.1"
    for comp in ("alloy", "sdlc"):
        config["metadataserver"][comp].update({"host": metadata_host, "port": metadata_port})
    config["relationalexecution"]["temporarytestdb"]["port"] = testdb_port

    config_path = os.path.join(tempfile.mkdtemp(prefix="pylegend_"), "config.json")
    with open(config_path, "w") as f:
        json.dump(config, f)
    return config_path


def _get_server_jar_path() -> Path:
    cache_dir = Path.home() / ".cache" / "pylegend"
    cache_dir.mkdir(parents=True, exist_ok=True)
    jar = cache_dir / f"legend-engine-server-http-server-{_LEGEND_ENGINE_VERSION}-shaded.jar"
    if not jar.exists():
        LOGGER.info("Downloading Legend server JAR %s...", _LEGEND_ENGINE_VERSION)
        try:
            with requests.get(_LEGEND_JAR_URL, stream=True) as r, open(jar, 'wb') as f:
                r.raise_for_status()
                for chunk in r.iter_content(8192):
                    f.write(chunk)
        except Exception as e:
            if jar.exists():
                jar.unlink()
            raise RuntimeError(f"Failed to download JAR: {e}") from e
    return jar


_singleton: PyLegendOptional[LocalLegendEnv] = None


def get_local_legend_env(**kwargs: object) -> LocalLegendEnv:
    global _singleton
    if _singleton is None:
        _singleton = LocalLegendEnv(**kwargs).start()  # type: ignore[arg-type]
        atexit.register(_singleton.stop)
    return _singleton


NORTHWIND_PROJECT_COORDINATES = VersionedProjectCoordinates(
    group_id="org.finos.legend.pylegend",
    artifact_id="pylegend-northwind-models",
    version="0.0.1-SNAPSHOT",
)



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import date, datetime
from decimal import Decimal as PythonDecimal

from pylegend._typing import (
    PyLegendList,
    PyLegendSequence,
    PyLegendDict,
    PyLegendCallable,
    PyLegendUnion,
)
from pylegend.core.language import (
    PyLegendPrimitive,
    PyLegendInteger,
    PyLegendFloat,
    PyLegendNumber,
    PyLegendBoolean,
    PyLegendString,
    PyLegendDate,
    PyLegendDateTime,
    PyLegendDecimal,
)
from pylegend.core.language.pandas_api.pandas_api_groupby_series import GroupbySeries
from pylegend.core.language.pandas_api.pandas_api_series import Series
from pylegend.core.language.pandas_api.pandas_api_tds_row import PandasApiTdsRow
from pylegend.core.language.shared.helpers import generate_pure_lambda
from pylegend.core.language.shared.literal_expressions import convert_literal_to_literal_expression
from pylegend.core.sql.metamodel import (
    QuerySpecification,
    SingleColumn,
)
from pylegend.core.tds.pandas_api.frames.functions.rank_function import RankFunction
from pylegend.core.tds.pandas_api.frames.helpers.series_helper import has_window_function
from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import PandasApiAppliedFunction
from pylegend.core.tds.pandas_api.frames.pandas_api_base_tds_frame import PandasApiBaseTdsFrame
from pylegend.core.tds.sql_query_helpers import copy_query, create_sub_query
from pylegend.core.tds.tds_column import TdsColumn, PrimitiveTdsColumn
from pylegend.core.tds.tds_frame import FrameToSqlConfig, FrameToPureConfig


class AssignFunction(PandasApiAppliedFunction):
    __base_frame: PandasApiBaseTdsFrame
    __col_definitions: PyLegendDict[
        str,
        PyLegendCallable[
            [PandasApiTdsRow],
            PyLegendUnion[int, float, bool, str, date, datetime, PythonDecimal, PyLegendPrimitive]
        ],
    ]

    @classmethod
    def name(cls) -> str:
        return "assign"  # pragma: no cover

    def __init__(
            self,
            base_frame: PandasApiBaseTdsFrame,
            col_definitions: PyLegendDict[
                str,
                PyLegendCallable[
                    [PandasApiTdsRow],
                    PyLegendUnion[int, float, bool, str, date, datetime, PythonDecimal, PyLegendPrimitive]
                ],
            ]
    ) -> None:
        self.__base_frame = base_frame
        self.__col_definitions = col_definitions

    def to_sql(self, config: FrameToSqlConfig) -> QuerySpecification:
        temp_column_name_suffix = "__INTERNAL_PYLEGEND_COLUMN__"
        db_extension = config.sql_to_string_generator().get_db_extension()
        base_query = self.__base_frame.to_sql_query_object(config)
        should_create_sub_query = (len(base_query.groupBy) > 0) or base_query.select.distinct

        new_query = (
            create_sub_query(base_query, config, "root") if should_create_sub_query else
            copy_query(base_query)
        )

        base_cols = {c.get_name() for c in self.__base_frame.columns()}
        tds_row = PandasApiTdsRow.from_tds_frame("c", self.__base_frame)
        for col, func in self.__col_definitions.items():
            res = func(tds_row)
            res_expr = res if isinstance(res, PyLegendPrimitive) else convert_literal_to_literal_expression(res)
            new_col_expr = res_expr.to_sql_expression(
                {"c": base_query},
                config
            )

            alias = db_extension.quote_identifier(col)
            if col in base_cols:
                for i, si in enumerate(new_query.select.selectItems):
                    if isinstance(si, SingleColumn) and si.alias == alias:
                        if isinstance(res, (Series, GroupbySeries)):
                            alias = (db_extension.quote_identifier(col + temp_column_name_suffix) if has_window_function(res)
                                     else alias)
                        new_query.select.selectItems[i] = SingleColumn(alias=alias, expression=new_col_expr)

            else:
                if isinstance(res, (Series, GroupbySeries)):
                    alias = (db_extension.quote_identifier(col + temp_column_name_suffix) if has_window_function(res)
                             else alias)
                new_query.select.selectItems.append(SingleColumn(alias=alias, expression=new_col_expr))

        expr_contains_window_func = False
        for col, func in self.__col_definitions.items():
            res = func(tds_row)
            if isinstance(res, (Series, GroupbySeries)):
                expr_contains_window_func |= has_window_function(res)

        if expr_contains_window_func:
            final_query = create_sub_query(new_query, config, "root")
            for col, func in self.__col_definitions.items():
                res = func(tds_row)
                if isinstance(res, (Series, GroupbySeries)):
                    if has_window_function(res):
                        alias = db_extension.quote_identifier(col + temp_column_name_suffix)
                        new_alias = db_extension.quote_identifier(col)
                        for i, si in enumerate(final_query.select.selectItems):
                            if isinstance(si, SingleColumn) and si.alias == alias:
                                final_query.select.selectItems[i] = SingleColumn(alias=new_alias, expression=si.expression)
            return final_query

        return new_query

    def to_pure(self, config: FrameToPureConfig) -> str:
        temp_column_name_suffix = "__INTERNAL_PYLEGEND_COLUMN__"
        tds_row = PandasApiTdsRow.from_tds_frame("c", self.__base_frame)
        base_cols = [c.get_name() for c in self.__base_frame.columns()]

        extend_exprs: PyLegendList[str] = []
        assigned_exprs: PyLegendDict[str, str] = {}
        for col, func in self.__col_definitions.items():
            res = func(tds_row)
            if isinstance(res, (Series, GroupbySeries)):
                sub_expressions = res.get_leaf_expressions()
                for expr in sub_expressions:
                    if isinstance(expr, Series):
                        applied_func = expr.get_filtered_frame().get_applied_function()
                    elif isinstance(expr, GroupbySeries):
                        applied_func = expr.raise_exception_if_no_function_applied().get_applied_function()
                    else:
                        continue

                    if isinstance(applied_func, RankFunction):
                        c, window = applied_func.construct_column_expression_and_window_tuples()[0]
                        window_expr = window.to_pure_expression(config)
                        function_expr = c[1].to_pure_expression(config)
                        target_col_name = c[0] + temp_column_name_suffix
                        extend = f"->extend({window_expr}, ~{target_col_name}:{generate_pure_lambda('p,w,r', function_expr)})"
                        extend_exprs.append(extend)
            res_expr = res if isinstance(res, PyLegendPrimitive) else convert_literal_to_literal_expression(res)
            assigned_exprs[col] = res_expr.to_pure_expression(config)

        # build project clauses
        clauses: PyLegendList[str] = []

        for col in base_cols:
            if col in assigned_exprs:
                clauses.append(f"{col}:c|{assigned_exprs[col]}")
            else:
                clauses.append(f"{col}:c|$c.{col}")

        for col, pure_expr in assigned_exprs.items():
            if col not in base_cols:
                clauses.append(f"{col}:c|{pure_expr}")

        return (
            f"{self.__base_frame.to_pure(config)}{config.separator(1)}"
            f"{config.separator(1).join(extend_exprs) + config.separator(1) if len(extend_exprs) > 0 else ''}"
            f"->project(~[{', '.join(clauses)}])"
        )

    def base_frame(self) -> PandasApiBaseTdsFrame:
        return self.__base_frame

    def tds_frame_parameters(self) -> PyLegendList["PandasApiBaseTdsFrame"]:
        return []

    def calculate_columns(self) -> PyLegendSequence["TdsColumn"]:
        new_cols = [c.copy() for c in self.__base_frame.columns() if c.get_name() not in self.__col_definitions]
        tds_row = PandasApiTdsRow.from_tds_frame("frame", self.__base_frame)
        for col, func in self.__col_definitions.items():
            res = func(tds_row)
            if isinstance(res, (int, PyLegendInteger)):
                new_cols.append(PrimitiveTdsColumn.integer_column(col))
            elif isinstance(res, (float, PyLegendFloat)):
                new_cols.append(PrimitiveTdsColumn.float_column(col))
            elif isinstance(res, (PythonDecimal, PyLegendDecimal)):
                new_cols.append(PrimitiveTdsColumn.decimal_column(col))
            elif isinstance(res, PyLegendNumber):
                new_cols.append(PrimitiveTdsColumn.number_column(col))  # pragma: no cover
            elif isinstance(res, (bool, PyLegendBoolean)):
                new_cols.append(
                    PrimitiveTdsColumn.boolean_column(col)
                )  # pragma: no cover (Boolean column not supported in PURE)
            elif isinstance(res, (str, PyLegendString)):
                new_cols.append(PrimitiveTdsColumn.string_column(col))
            elif isinstance(res, (datetime, PyLegendDateTime)):
                new_cols.append(PrimitiveTdsColumn.datetime_column(col))
            elif isinstance(res, (date, PyLegendDate)):
                new_cols.append(PrimitiveTdsColumn.date_column(col))
            else:
                raise RuntimeError("Type not supported")
        return new_cols

    def validate(self) -> bool:
        tds_row = PandasApiTdsRow.from_tds_frame("frame", self.__base_frame)
        for col, f in self.__col_definitions.items():
            f(tds_row)
        return True



# Copyright 2026 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendSequence,
)
import builtins
from pylegend.core.tds.tds_column import PrimitiveType
from pylegend.core.tds.cast_helpers import CastTarget

__all__: PyLegendSequence[str] = [
    # Simple types
    "boolean",
    "string",
    "integer",
    "float",
    "number",
    "decimal",
    "date",
    "datetime",
    "strictdate",

    # Precise integer types
    "tinyint",
    "utinyint",
    "smallint",
    "usmallint",
    "int",
    "uint",
    "bigint",
    "ubigint",

    # Precise float types
    "float4",
    "double",

    # Precise date types
    "timestamp",

    # Parameterized types
    "varchar",
    "numeric",
]


def boolean() -> CastTarget:
    """Cast to Boolean."""
    return PrimitiveType.Boolean


def string() -> CastTarget:
    """Cast to String."""
    return PrimitiveType.String


def integer() -> CastTarget:
    """Cast to Integer."""
    return PrimitiveType.Integer


def float() -> CastTarget:
    """Cast to Float."""
    return PrimitiveType.Float


def number() -> CastTarget:
    """Cast to Number."""
    return PrimitiveType.Number


def decimal() -> CastTarget:
    """Cast to Decimal."""
    return PrimitiveType.Decimal


def date() -> CastTarget:
    """Cast to Date."""
    return PrimitiveType.Date  # pragma: no cover


def datetime() -> CastTarget:
    """Cast to DateTime."""
    return PrimitiveType.DateTime


def strictdate() -> CastTarget:
    """Cast to StrictDate."""
    return PrimitiveType.StrictDate


def tinyint() -> CastTarget:
    """Cast to TinyInt."""
    return PrimitiveType.TinyInt


def utinyint() -> CastTarget:
    """Cast to UTinyInt."""
    return PrimitiveType.UTinyInt


def smallint() -> CastTarget:
    """Cast to SmallInt."""
    return PrimitiveType.SmallInt


def usmallint() -> CastTarget:
    """Cast to USmallInt."""
    return PrimitiveType.USmallInt


def int() -> CastTarget:
    """Cast to Int."""
    return PrimitiveType.Int


def uint() -> CastTarget:
    """Cast to UInt."""
    return PrimitiveType.UInt


def bigint() -> CastTarget:
    """Cast to BigInt."""
    return PrimitiveType.BigInt


def ubigint() -> CastTarget:
    """Cast to UBigInt."""
    return PrimitiveType.UBigInt


def float4() -> CastTarget:
    """Cast to Float4."""
    return PrimitiveType.Float4


def double() -> CastTarget:
    """Cast to Double."""
    return PrimitiveType.Double


def timestamp() -> CastTarget:
    """Cast to Timestamp."""
    return PrimitiveType.Timestamp


def varchar(max_length: builtins.int) -> CastTarget:  # noqa: F821
    """Cast to Varchar with a given max length.

    Args:
        max_length: Maximum character length for the Varchar column.

    Example::

        frame.cast({"name": pylegend.type_factory.varchar(200)})
    """
    return (PrimitiveType.Varchar, max_length)  # type: ignore


def numeric(precision: builtins.int, scale: builtins.int) -> CastTarget:  # noqa: F821
    """Cast to Numeric with given precision and scale.

    Args:
        precision: Total number of digits.
        scale: Number of digits after the decimal point.

    Example::

        frame.cast({"amount": pylegend.type_factory.numeric(10, 2)})
    """
    return (PrimitiveType.Numeric, precision, scale)  # type: ignore



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendSequence,
    PyLegendDict,
    PyLegendUnion,
    TYPE_CHECKING,
)
from pylegend.core.language.shared.primitives.number import PyLegendNumber
from pylegend.core.language.shared.expression import PyLegendExpressionFloatReturn
from pylegend.core.language.shared.literal_expressions import PyLegendFloatLiteralExpression
from pylegend.core.sql.metamodel import (
    Expression,
    QuerySpecification
)
from pylegend.core.tds.pandas_api.frames.helpers.series_helper import grammar_method
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.core.language.shared.operations.float_operation_expressions import (
    PyLegendFloatAbsoluteExpression,
    PyLegendFloatAddExpression,
    PyLegendFloatNegativeExpression,
    PyLegendFloatSubtractExpression,
    PyLegendFloatMultiplyExpression,
)
if TYPE_CHECKING:
    from pylegend.core.language.shared.primitives import PyLegendInteger

__all__: PyLegendSequence[str] = [
    "PyLegendFloat"
]


class PyLegendFloat(PyLegendNumber):
    __value_copy: PyLegendExpressionFloatReturn

    def __init__(
            self,
            value: PyLegendExpressionFloatReturn
    ) -> None:
        self.__value_copy = value
        super().__init__(value)

    @grammar_method
    def __add__(
            self,
            other: PyLegendUnion[int, float, "PyLegendInteger", "PyLegendFloat", "PyLegendNumber"]
    ) -> "PyLegendUnion[PyLegendNumber, PyLegendFloat]":
        PyLegendNumber.validate_param_to_be_number(other, "Float plus (+) parameter")
        if isinstance(other, (float, PyLegendFloat)):
            other_op = PyLegendFloat.__convert_to_float_expr(other)
            return PyLegendFloat(PyLegendFloatAddExpression(self.__value_copy, other_op))
        else:
            return super().__add__(other)

    @grammar_method
    def __radd__(
            self,
            other: PyLegendUnion[int, float, "PyLegendInteger", "PyLegendFloat", "PyLegendNumber"]
    ) -> "PyLegendUnion[PyLegendNumber, PyLegendFloat]":
        PyLegendNumber.validate_param_to_be_number(other, "Float plus (+) parameter")
        if isinstance(other, (float, PyLegendFloat)):
            other_op = PyLegendFloat.__convert_to_float_expr(other)
            return PyLegendFloat(PyLegendFloatAddExpression(other_op, self.__value_copy))
        else:
            return super().__radd__(other)

    @grammar_method
    def __sub__(
            self,
            other: PyLegendUnion[int, float, "PyLegendInteger", "PyLegendFloat", "PyLegendNumber"]
    ) -> "PyLegendUnion[PyLegendNumber, PyLegendFloat]":
        PyLegendNumber.validate_param_to_be_number(other, "Float minus (-) parameter")
        if isinstance(other, (float, PyLegendFloat)):
            other_op = PyLegendFloat.__convert_to_float_expr(other)
            return PyLegendFloat(PyLegendFloatSubtractExpression(self.__value_copy, other_op))
        else:
            return super().__sub__(other)

    @grammar_method
    def __rsub__(
            self,
            other: PyLegendUnion[int, float, "PyLegendInteger", "PyLegendFloat", "PyLegendNumber"]
    ) -> "PyLegendUnion[PyLegendNumber, PyLegendFloat]":
        PyLegendNumber.validate_param_to_be_number(other, "Float minus (-) parameter")
        if isinstance(other, (float, PyLegendFloat)):
            other_op = PyLegendFloat.__convert_to_float_expr(other)
            return PyLegendFloat(PyLegendFloatSubtractExpression(other_op, self.__value_copy))
        else:
            return super().__rsub__(other)

    @grammar_method
    def __mul__(
            self,
            other: PyLegendUnion[int, float, "PyLegendInteger", "PyLegendFloat", "PyLegendNumber"]
    ) -> "PyLegendUnion[PyLegendNumber, PyLegendFloat]":
        PyLegendNumber.validate_param_to_be_number(other, "Float multiply (*) parameter")
        if isinstance(other, (float, PyLegendFloat)):
            other_op = PyLegendFloat.__convert_to_float_expr(other)
            return PyLegendFloat(PyLegendFloatMultiplyExpression(self.__value_copy, other_op))
        else:
            return super().__mul__(other)

    @grammar_method
    def __rmul__(
            self,
            other: PyLegendUnion[int, float, "PyLegendInteger", "PyLegendFloat", "PyLegendNumber"]
    ) -> "PyLegendUnion[PyLegendNumber, PyLegendFloat]":
        PyLegendNumber.validate_param_to_be_number(other, "Float multiply (*) parameter")
        if isinstance(other, (float, PyLegendFloat)):
            other_op = PyLegendFloat.__convert_to_float_expr(other)
            return PyLegendFloat(PyLegendFloatMultiplyExpression(other_op, self.__value_copy))
        else:
            return super().__rmul__(other)

    @grammar_method
    def __abs__(self) -> "PyLegendFloat":
        return PyLegendFloat(PyLegendFloatAbsoluteExpression(self.__value_copy))

    @grammar_method
    def __neg__(self) -> "PyLegendFloat":
        return PyLegendFloat(PyLegendFloatNegativeExpression(self.__value_copy))

    @grammar_method
    def __pos__(self) -> "PyLegendFloat":
        return self

    @staticmethod
    def __convert_to_float_expr(
            val: PyLegendUnion[float, "PyLegendFloat"]
    ) -> PyLegendExpressionFloatReturn:
        if isinstance(val, float):
            return PyLegendFloatLiteralExpression(val)
        return val.__value_copy

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return super().to_sql_expression(frame_name_to_base_query_map, config)

    def value(self) -> PyLegendExpressionFloatReturn:
        return self.__value_copy

    @staticmethod
    def __validate__param_to_be_float(params, desc):  # type: ignore
        PyLegendNumber.validate_param_to_be_number(params, desc)



# Copyright 2026 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import importlib
from textwrap import dedent
from typing import (
    Type,
    TypeVar,
)
from pylegend._typing import (
    PyLegendAny,
    PyLegendCallable,
    PyLegendList,
    PyLegendOptional,
    PyLegendSequence,
    PyLegendUnion,
    TYPE_CHECKING,
)
from pylegend.core.language.shared.expression import PyLegendExpression
if TYPE_CHECKING:
    from pylegend.core.language.pandas_api.pandas_api_series import Series
    from pylegend.core.language.pandas_api.pandas_api_groupby_series import GroupbySeries


__all__: PyLegendSequence[str] = [
    "add_primitive_methods",
    "assert_and_find_core_series",
    "has_window_function",
]

T = TypeVar("T")
F = TypeVar("F", bound=PyLegendCallable[..., PyLegendAny])  # type: ignore[explicit-any]


def grammar_method(func: F) -> F:  # type: ignore[explicit-any]
    setattr(func, "_is_grammar_method", True)
    return func


def add_primitive_methods(cls: Type[T]) -> Type[T]:
    primitive_to_series_map = {
        "PyLegendBoolean": "BooleanSeries",
        "PyLegendString": "StringSeries",
        "PyLegendNumber": "NumberSeries",
        "PyLegendInteger": "IntegerSeries",
        "PyLegendFloat": "FloatSeries",
        "PyLegendDate": "DateSeries",
        "PyLegendDateTime": "DateTimeSeries",
        "PyLegendStrictDate": "StrictDateSeries"
    }

    primitive_to_groupby_series_map = {
        "PyLegendBoolean": "BooleanGroupbySeries",
        "PyLegendString": "StringGroupbySeries",
        "PyLegendNumber": "NumberGroupbySeries",
        "PyLegendInteger": "IntegerGroupbySeries",
        "PyLegendFloat": "FloatGroupbySeries",
        "PyLegendDate": "DateGroupbySeries",
        "PyLegendDateTime": "DateTimeGroupbySeries",
        "PyLegendStrictDate": "StrictDateGroupbySeries"
    }

    mro_names = [base.__name__ for base in cls.__mro__]

    series_type: str
    if "GroupbySeries" in mro_names:
        series_type = "GroupbySeries"
    elif "Series" in mro_names:
        series_type = "Series"
    else:
        raise NotImplementedError(f"Can't add primitive methods to class of type: {cls.__name__}")  # pragma: no cover

    if series_type == "GroupbySeries":
        target_map = primitive_to_groupby_series_map
        target_module_path = "pylegend.core.language.pandas_api.pandas_api_groupby_series"
    else:
        target_map = primitive_to_series_map
        target_module_path = "pylegend.core.language.pandas_api.pandas_api_series"

    methods_to_wrap = {}
    for base in cls.__mro__:
        for name, attr in base.__dict__.items():
            if name not in methods_to_wrap:
                if callable(attr) and getattr(attr, "_is_grammar_method", False):
                    methods_to_wrap[name] = attr

    for name, original_func in methods_to_wrap.items():
        if name in cls.__dict__ and not getattr(cls.__dict__[name], "_is_grammar_method", False):  # pragma: no cover
            continue

        def make_wrapper(  # type: ignore[explicit-any]
                func: PyLegendCallable[..., PyLegendAny]
        ) -> PyLegendCallable[..., PyLegendAny]:
            def wrapper(  # type: ignore[explicit-any]
                    self: PyLegendAny,
                    *args: PyLegendAny,
                    **kwargs: PyLegendAny
            ) -> PyLegendAny:
                result_primitive = func(self, *args, **kwargs)
                primitive_type_name = type(result_primitive).__name__

                if not hasattr(result_primitive, 'value') or primitive_type_name not in target_map:
                    return result_primitive  # pragma: no cover

                base_frame = self.get_base_frame()
                col_name = self.columns()[0].get_name()

                target_class_str = target_map[primitive_type_name]
                module = importlib.import_module(target_module_path)
                TargetSeriesClass = getattr(module, target_class_str)

                expr = result_primitive.value()
                if series_type == "GroupbySeries":
                    return TargetSeriesClass(base_frame, None, expr)
                else:
                    return TargetSeriesClass(base_frame, col_name, expr)

            return wrapper

        setattr(cls, name, make_wrapper(original_func))

    return cls


def assert_and_find_core_series(expr: PyLegendExpression) -> PyLegendOptional[PyLegendUnion["Series", "GroupbySeries"]]:
    from pylegend.core.language.pandas_api.pandas_api_series import Series
    from pylegend.core.language.pandas_api.pandas_api_groupby_series import GroupbySeries

    core_series_list: PyLegendList[PyLegendUnion[Series, GroupbySeries]] = []

    sub_expressions = expr.get_leaf_expressions()
    for expr in sub_expressions:
        if isinstance(expr, (Series, GroupbySeries)):
            core_series_list.append(expr)

    if len(core_series_list) == 0:
        return None  # pragma: no cover
    elif len(core_series_list) == 1:
        return core_series_list[0]
    else:
        core_series_with_applied_function = [series for series in core_series_list if series.has_applied_function()]
        if len(core_series_with_applied_function) == 0:
            return core_series_list[0]
        elif len(core_series_with_applied_function) == 1:
            return core_series_with_applied_function[0]

        error_msg = '''
            Only expressions with maximum one Series/GroupbySeries function call (such as .rank()) is supported.
            If multiple Series/GroupbySeries need function calls, please compute them in separate steps.
            For example,
                unsupported:
                    frame['new_col'] = frame['col1'].rank() + 2 + frame['col2'].rank()
                supported:
                    frame['new_col'] = frame['col1'].rank() + 2
                    frame['new_col'] += frame['col2'].rank()
        '''
        error_msg = dedent(error_msg).strip()
        raise ValueError(error_msg)


def has_window_function(series: PyLegendUnion["Series", "GroupbySeries"]) -> bool:
    from pylegend.core.language.pandas_api.pandas_api_series import Series
    from pylegend.core.language.pandas_api.pandas_api_groupby_series import GroupbySeries
    from pylegend.core.tds.pandas_api.frames.functions.rank_function import RankFunction

    if series.expr is not None:
        core_series = assert_and_find_core_series(series.expr)
        assert core_series is not None
        return has_window_function(core_series)

    considered_window_functions = [RankFunction]

    if isinstance(series, Series):
        applied_func = series.get_filtered_frame().get_applied_function()
        return any(isinstance(applied_func, window_function) for window_function in considered_window_functions)

    elif isinstance(series, GroupbySeries):
        applied_func = series.raise_exception_if_no_function_applied().get_applied_function()
        return any(isinstance(applied_func, window_function) for window_function in considered_window_functions)

    else:
        raise TypeError("Window function's existence can only be checked in a Series or a GroupbySeries")  # pragma: no cover



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import date, datetime
from pylegend._typing import (
    PyLegendSequence,
    PyLegendDict,
    PyLegendUnion,
    TYPE_CHECKING,
)
from pylegend.core.language.shared.primitives.primitive import PyLegendPrimitive
from pylegend.core.language.shared.primitives.integer import PyLegendInteger
from pylegend.core.language.shared.primitives.boolean import PyLegendBoolean
from pylegend.core.language.shared.expression import (
    PyLegendExpressionDateReturn,
)
from pylegend.core.language.shared.literal_expressions import (
    PyLegendDateTimeLiteralExpression,
    PyLegendStrictDateLiteralExpression,
    PyLegendIntegerLiteralExpression,
    PyLegendStringLiteralExpression,
)
from pylegend.core.language.shared.operations.date_operation_expressions import (
    PyLegendFirstDayOfYearExpression,
    PyLegendFirstDayOfQuarterExpression,
    PyLegendFirstDayOfMonthExpression,
    PyLegendFirstDayOfWeekExpression,
    PyLegendFirstHourOfDayExpression,
    PyLegendFirstMinuteOfHourExpression,
    PyLegendFirstSecondOfMinuteExpression,
    PyLegendFirstMillisecondOfSecondExpression,
    PyLegendYearExpression,
    PyLegendQuarterExpression,
    PyLegendMonthExpression,
    PyLegendWeekOfYearExpression,
    PyLegendDayOfYearExpression,
    PyLegendDayOfMonthExpression,
    PyLegendDayOfWeekExpression,
    PyLegendHourExpression,
    PyLegendMinuteExpression,
    PyLegendSecondExpression,
    PyLegendEpochExpression,
    PyLegendDatePartExpression,
    PyLegendDateLessThanExpression,
    PyLegendDateLessThanEqualExpression,
    PyLegendDateGreaterThanExpression,
    PyLegendDateGreaterThanEqualExpression,
    PyLegendDateAdjustExpression,
    PyLegendDateDiffExpression,
)
from pylegend.core.sql.metamodel import (
    Expression,
    QuerySpecification
)
from pylegend.core.tds.pandas_api.frames.helpers.series_helper import grammar_method
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.core.tds.tds_frame import FrameToPureConfig
if TYPE_CHECKING:
    from pylegend.core.language.shared.primitives.datetime import PyLegendDateTime
    from pylegend.core.language.shared.primitives.strictdate import PyLegendStrictDate


__all__: PyLegendSequence[str] = [
    "PyLegendDate"
]


class PyLegendDate(PyLegendPrimitive):
    __value: PyLegendExpressionDateReturn

    def __init__(
            self,
            value: PyLegendExpressionDateReturn
    ) -> None:
        self.__value = value

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return self.__value.to_sql_expression(frame_name_to_base_query_map, config)

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        return self.__value.to_pure_expression(config)

    def value(self) -> PyLegendExpressionDateReturn:
        return self.__value

    @grammar_method
    def first_day_of_year(self) -> "PyLegendDate":
        return PyLegendDate(PyLegendFirstDayOfYearExpression(self.__value))

    @grammar_method
    def first_day_of_quarter(self) -> "PyLegendDate":
        return PyLegendDate(PyLegendFirstDayOfQuarterExpression(self.__value))

    @grammar_method
    def first_day_of_month(self) -> "PyLegendDate":
        return PyLegendDate(PyLegendFirstDayOfMonthExpression(self.__value))

    @grammar_method
    def first_day_of_week(self) -> "PyLegendDate":
        return PyLegendDate(PyLegendFirstDayOfWeekExpression(self.__value))

    @grammar_method
    def first_hour_of_day(self) -> "PyLegendDateTime":
        from pylegend.core.language.shared.primitives.datetime import PyLegendDateTime
        return PyLegendDateTime(PyLegendFirstHourOfDayExpression(self.__value))

    @grammar_method
    def first_minute_of_hour(self) -> "PyLegendDateTime":
        from pylegend.core.language.shared.primitives.datetime import PyLegendDateTime
        return PyLegendDateTime(PyLegendFirstMinuteOfHourExpression(self.__value))

    @grammar_method
    def first_second_of_minute(self) -> "PyLegendDateTime":
        from pylegend.core.language.shared.primitives.datetime import PyLegendDateTime
        return PyLegendDateTime(PyLegendFirstSecondOfMinuteExpression(self.__value))

    @grammar_method
    def first_millisecond_of_second(self) -> "PyLegendDateTime":
        from pylegend.core.language.shared.primitives.datetime import PyLegendDateTime
        return PyLegendDateTime(PyLegendFirstMillisecondOfSecondExpression(self.__value))

    @grammar_method
    def year(self) -> "PyLegendInteger":
        return PyLegendInteger(PyLegendYearExpression(self.__value))

    @grammar_method
    def month(self) -> "PyLegendInteger":
        return PyLegendInteger(PyLegendMonthExpression(self.__value))

    @grammar_method
    def day(self) -> "PyLegendInteger":
        return PyLegendInteger(PyLegendDayOfMonthExpression(self.__value))

    @grammar_method
    def hour(self) -> "PyLegendInteger":
        return PyLegendInteger(PyLegendHourExpression(self.__value))

    @grammar_method
    def minute(self) -> "PyLegendInteger":
        return PyLegendInteger(PyLegendMinuteExpression(self.__value))

    @grammar_method
    def second(self) -> "PyLegendInteger":
        return PyLegendInteger(PyLegendSecondExpression(self.__value))

    @grammar_method
    def epoch_value(self) -> "PyLegendInteger":
        return PyLegendInteger(PyLegendEpochExpression(self.__value))

    @grammar_method
    def quarter(self) -> "PyLegendInteger":
        return PyLegendInteger(PyLegendQuarterExpression(self.__value))

    @grammar_method
    def week_of_year(self) -> "PyLegendInteger":
        return PyLegendInteger(PyLegendWeekOfYearExpression(self.__value))

    @grammar_method
    def day_of_year(self) -> "PyLegendInteger":
        return PyLegendInteger(PyLegendDayOfYearExpression(self.__value))

    @grammar_method
    def day_of_week(self) -> "PyLegendInteger":
        return PyLegendInteger(PyLegendDayOfWeekExpression(self.__value))

    @grammar_method
    def date_part(self) -> "PyLegendStrictDate":
        from pylegend.core.language.shared.primitives.strictdate import PyLegendStrictDate
        return PyLegendStrictDate(PyLegendDatePartExpression(self.__value))

    @grammar_method
    def timedelta(self, number: PyLegendUnion[int, "PyLegendInteger"], duration_unit: str) -> "PyLegendDate":
        self.validate_param_to_be_int_or_int_expr(number, "timedelta number parameter")
        number_op = PyLegendIntegerLiteralExpression(number) if isinstance(number, int) else number.value()
        self.validate_duration_unit_param(duration_unit)
        duration_unit_op = PyLegendStringLiteralExpression(duration_unit.upper())
        return PyLegendDate(PyLegendDateAdjustExpression([self.__value, number_op, duration_unit_op]))

    @grammar_method
    def diff(
            self,
            other: PyLegendUnion[date, datetime, "PyLegendStrictDate", "PyLegendDateTime", "PyLegendDate"],
            duration_unit: str) -> "PyLegendInteger":
        self.validate_param_to_be_date(other, "Diff parameter")
        other_op = PyLegendDate.__convert_to_date_expr(other)
        self.validate_duration_unit_param(duration_unit)
        duration_unit_op = PyLegendStringLiteralExpression(duration_unit.upper())
        return PyLegendInteger(PyLegendDateDiffExpression([self.__value, other_op, duration_unit_op]))

    @grammar_method
    def __lt__(
            self,
            other: PyLegendUnion[date, datetime, "PyLegendStrictDate", "PyLegendDateTime", "PyLegendDate"]
    ) -> "PyLegendBoolean":
        PyLegendDate.validate_param_to_be_date(other, "Date less than (<) parameter")
        other_op = PyLegendDate.__convert_to_date_expr(other)
        return PyLegendBoolean(PyLegendDateLessThanExpression(self.__value, other_op))

    @grammar_method
    def __le__(
            self,
            other: PyLegendUnion[date, datetime, "PyLegendStrictDate", "PyLegendDateTime", "PyLegendDate"]
    ) -> "PyLegendBoolean":
        PyLegendDate.validate_param_to_be_date(other, "Date less than equal (<=) parameter")
        other_op = PyLegendDate.__convert_to_date_expr(other)
        return PyLegendBoolean(PyLegendDateLessThanEqualExpression(self.__value, other_op))

    @grammar_method
    def __gt__(
            self,
            other: PyLegendUnion[date, datetime, "PyLegendStrictDate", "PyLegendDateTime", "PyLegendDate"]
    ) -> "PyLegendBoolean":
        PyLegendDate.validate_param_to_be_date(other, "Date greater than (>) parameter")
        other_op = PyLegendDate.__convert_to_date_expr(other)
        return PyLegendBoolean(PyLegendDateGreaterThanExpression(self.__value, other_op))

    @grammar_method
    def __ge__(
            self,
            other: PyLegendUnion[date, datetime, "PyLegendStrictDate", "PyLegendDateTime", "PyLegendDate"]
    ) -> "PyLegendBoolean":
        PyLegendDate.validate_param_to_be_date(other, "Date greater than equal (>=) parameter")
        other_op = PyLegendDate.__convert_to_date_expr(other)
        return PyLegendBoolean(PyLegendDateGreaterThanEqualExpression(self.__value, other_op))

    @staticmethod
    def __convert_to_date_expr(
            val: PyLegendUnion[date, datetime, "PyLegendDateTime", "PyLegendStrictDate", "PyLegendDate"]
    ) -> PyLegendExpressionDateReturn:
        if isinstance(val, datetime):
            return PyLegendDateTimeLiteralExpression(val)
        if isinstance(val, date):
            return PyLegendStrictDateLiteralExpression(val)
        return val.__value

    @staticmethod
    def validate_param_to_be_date(
            param: PyLegendUnion[date, datetime, "PyLegendDateTime", "PyLegendStrictDate", "PyLegendDate"],
            desc: str
    ) -> None:
        from pylegend.core.language.shared.primitives.datetime import PyLegendDateTime
        from pylegend.core.language.shared.primitives.strictdate import PyLegendStrictDate
        if not isinstance(param, (date, datetime, PyLegendDateTime, PyLegendStrictDate, PyLegendDate)):
            raise TypeError(desc + " should be a datetime.date/datetime.datetime or a Date expression"
                                   " (PyLegendDateTime/PyLegendStrictDate/PyLegendDate)."
                                   " Got value " + str(param) + " of type: " + str(type(param)))

    @staticmethod
    def validate_duration_unit_param(duration_unit: str) -> None:
        if duration_unit.lower() not in ('years', 'months', 'weeks', 'days', 'hours', 'minutes', 'seconds',
                                         'milliseconds', 'microseconds', 'nanoseconds'):
            raise ValueError(
                f"Unknown duration unit - {duration_unit}. Supported values are - YEARS, MONTHS, WEEKS, DAYS, HOURS,"
                " MINUTES, SECONDS, MILLISECONDS, MICROSECONDS, NANOSECONDS"
            )

    @staticmethod
    def validate_param_to_be_int_or_int_expr(
            param: PyLegendUnion[int, "PyLegendInteger"],
            desc: str
    ) -> None:
        if not isinstance(param, (int, PyLegendInteger)):
            raise TypeError(desc + " should be a int or an integer expression (PyLegendInteger)."
                                   " Got value " + str(param) + " of type: " + str(type(param)))



# Copyright 2026 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

import pandas as pd

from pylegend._typing import (
    PyLegendUnion,
    PyLegendSequence,
    PyLegendTuple,
    PyLegendCallable
)
from pylegend.core.language import PyLegendBoolean
from pylegend.core.tds.pandas_api.frames.functions.filtering import PandasApiFilteringFunction
from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import PandasApiAppliedFunctionTdsFrame

if TYPE_CHECKING:
    from pylegend.core.tds.pandas_api.frames.pandas_api_base_tds_frame import PandasApiBaseTdsFrame
    from pylegend.core.tds.pandas_api.frames.pandas_api_tds_frame import PandasApiTdsFrame

__all__: PyLegendSequence[str] = [
    "PandasApiLocIndexer"
]


class PandasApiLocIndexer:
    _frame: "PandasApiBaseTdsFrame"

    def __init__(self, frame: "PandasApiBaseTdsFrame") -> None:
        self._frame = frame

    def __getitem__(  # type: ignore
            self,
            key: PyLegendUnion[
                slice,
                PyLegendBoolean,
                PyLegendCallable[["PandasApiBaseTdsFrame"], PyLegendBoolean],
                PyLegendTuple[
                    PyLegendUnion[slice, PyLegendBoolean, PyLegendCallable[["PandasApiBaseTdsFrame"], PyLegendBoolean]],
                    PyLegendUnion[str, slice, PyLegendSequence[str], PyLegendSequence[bool]]
                ]
            ]
    ) -> "PandasApiTdsFrame":
        rows: PyLegendUnion[  # type: ignore
            slice,
            PyLegendBoolean,
            PyLegendCallable[["PandasApiBaseTdsFrame"], PyLegendBoolean]
        ]
        cols: PyLegendUnion[str, slice, PyLegendSequence[str], PyLegendSequence[bool]]  # type: ignore

        if isinstance(key, tuple):
            if len(key) == 1:
                rows, cols = key[0], slice(None, None, None)
            elif len(key) == 2:
                rows, cols = key[0], key[1]
            else:
                raise IndexError("Too many indexers")
        else:
            rows, cols = key, slice(None, None, None)

        row_frame = self._handle_row_selection(rows)
        return self._handle_column_selection(row_frame, cols)

    def _handle_row_selection(  # type: ignore
            self,
            rows: PyLegendUnion[slice, PyLegendBoolean, PyLegendCallable[["PandasApiBaseTdsFrame"], PyLegendBoolean]]
    ) -> "PandasApiTdsFrame":
        if isinstance(rows, slice):
            if rows.start is None and rows.stop is None and rows.step is None:
                return self._frame
            else:
                raise TypeError(
                    "loc supports only ':' for row slicing. "
                    "Label-based slicing for rows is not supported."
                )

        if isinstance(rows, PyLegendBoolean):
            return PandasApiAppliedFunctionTdsFrame(
                PandasApiFilteringFunction(self._frame, filter_expr=rows)
            )

        if callable(rows):
            new_key = rows(self._frame)
            return self._handle_row_selection(new_key)

        raise TypeError(f"Unsupported key type for .loc row selection: {type(rows)}")

    def _handle_column_selection(  # type: ignore
            self,
            frame: "PandasApiTdsFrame",
            cols: PyLegendUnion[str, slice, PyLegendSequence[str], PyLegendSequence[bool]]
    ) -> "PandasApiTdsFrame":
        if isinstance(cols, slice) and cols.start is None and cols.stop is None and cols.step is None:
            return frame

        if isinstance(cols, str):
            return frame.filter(items=[cols])

        if isinstance(cols, (list, tuple)):
            all_columns = [c.get_name() for c in frame.columns()]
            is_boolean_list = all(isinstance(k, bool) for k in cols)

            if is_boolean_list:
                if len(cols) != len(all_columns):
                    raise IndexError(f"Boolean index has wrong length: {len(cols)} instead of {len(all_columns)}")
                selected_columns = [col for col, select in zip(all_columns, cols) if select]
                return frame.filter(items=selected_columns)
            else:
                missing_cols = [c for c in cols if c not in all_columns]
                if missing_cols:
                    raise KeyError(f"{missing_cols} not in index")
                return frame.filter(items=cols)  # type: ignore

        if isinstance(cols, slice):
            all_columns = [c.get_name() for c in frame.columns()]
            pd_index = pd.Index(all_columns)

            slicer = pd_index.slice_indexer(start=cols.start, end=cols.stop, step=cols.step)
            selected_columns = pd_index[slicer].tolist()
            if not selected_columns:
                return frame.head(0)
            return frame.filter(items=selected_columns)

        raise TypeError(f"Unsupported key type for .loc column selection: {type(cols)}")



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendSequence,
    PyLegendDict,
)
from pylegend.core.language.shared.expression import (
    PyLegendExpressionBooleanReturn,
)
from pylegend.core.language.shared.helpers import generate_pure_functional_call
from pylegend.core.language.shared.operations.binary_expression import PyLegendBinaryExpression
from pylegend.core.language.shared.operations.unary_expression import PyLegendUnaryExpression
from pylegend.core.sql.metamodel import (
    Expression,
    QuerySpecification,
    LogicalBinaryExpression,
    LogicalBinaryType,
    NotExpression,
    ComparisonExpression,
    ComparisonOperator,
)
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.core.tds.tds_frame import FrameToPureConfig


__all__: PyLegendSequence[str] = [
    "PyLegendBooleanOrExpression",
    "PyLegendBooleanAndExpression",
    "PyLegendBooleanNotExpression",
    "PyLegendBooleanLessThanExpression",
    "PyLegendBooleanLessThanEqualExpression",
    "PyLegendBooleanGreaterThanExpression",
    "PyLegendBooleanGreaterThanEqualExpression",
    "PyLegendBooleanXorExpression"
]


class PyLegendBooleanOrExpression(PyLegendBinaryExpression, PyLegendExpressionBooleanReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return LogicalBinaryExpression(LogicalBinaryType.OR, expression1, expression2)

    @staticmethod
    def __to_pure_func(op1_expr: str, op2_expr: str, config: FrameToPureConfig) -> str:
        return f"({op1_expr} || {op2_expr})"

    def __init__(self, operand1: PyLegendExpressionBooleanReturn, operand2: PyLegendExpressionBooleanReturn) -> None:
        PyLegendExpressionBooleanReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendBooleanOrExpression.__to_sql_func,
            PyLegendBooleanOrExpression.__to_pure_func,
            non_nullable=True,
            first_operand_needs_to_be_non_nullable=True,
            second_operand_needs_to_be_non_nullable=True
        )


class PyLegendBooleanAndExpression(PyLegendBinaryExpression, PyLegendExpressionBooleanReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return LogicalBinaryExpression(LogicalBinaryType.AND, expression1, expression2)

    @staticmethod
    def __to_pure_func(op1_expr: str, op2_expr: str, config: FrameToPureConfig) -> str:
        return f"({op1_expr} && {op2_expr})"

    def __init__(self, operand1: PyLegendExpressionBooleanReturn, operand2: PyLegendExpressionBooleanReturn) -> None:
        PyLegendExpressionBooleanReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendBooleanAndExpression.__to_sql_func,
            PyLegendBooleanAndExpression.__to_pure_func,
            non_nullable=True,
            first_operand_needs_to_be_non_nullable=True,
            second_operand_needs_to_be_non_nullable=True
        )


class PyLegendBooleanLessThanExpression(PyLegendBinaryExpression, PyLegendExpressionBooleanReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return ComparisonExpression(expression1, expression2, ComparisonOperator.LESS_THAN)

    @staticmethod
    def __to_pure_func(op1_expr: str, op2_expr: str, config: FrameToPureConfig) -> str:
        return f"({op1_expr} < {op2_expr})"

    def __init__(self, operand1: PyLegendExpressionBooleanReturn, operand2: PyLegendExpressionBooleanReturn) -> None:
        PyLegendExpressionBooleanReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendBooleanLessThanExpression.__to_sql_func,
            PyLegendBooleanLessThanExpression.__to_pure_func,
            non_nullable=True
        )


class PyLegendBooleanLessThanEqualExpression(PyLegendBinaryExpression, PyLegendExpressionBooleanReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return ComparisonExpression(expression1, expression2, ComparisonOperator.LESS_THAN_OR_EQUAL)

    @staticmethod
    def __to_pure_func(op1_expr: str, op2_expr: str, config: FrameToPureConfig) -> str:
        return f"({op1_expr} <= {op2_expr})"

    def __init__(self, operand1: PyLegendExpressionBooleanReturn, operand2: PyLegendExpressionBooleanReturn) -> None:
        PyLegendExpressionBooleanReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendBooleanLessThanEqualExpression.__to_sql_func,
            PyLegendBooleanLessThanEqualExpression.__to_pure_func,
            non_nullable=True
        )


class PyLegendBooleanGreaterThanExpression(PyLegendBinaryExpression, PyLegendExpressionBooleanReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return ComparisonExpression(expression1, expression2, ComparisonOperator.GREATER_THAN)

    @staticmethod
    def __to_pure_func(op1_expr: str, op2_expr: str, config: FrameToPureConfig) -> str:
        return f"({op1_expr} > {op2_expr})"

    def __init__(self, operand1: PyLegendExpressionBooleanReturn, operand2: PyLegendExpressionBooleanReturn) -> None:
        PyLegendExpressionBooleanReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendBooleanGreaterThanExpression.__to_sql_func,
            PyLegendBooleanGreaterThanExpression.__to_pure_func,
            non_nullable=True
        )


class PyLegendBooleanGreaterThanEqualExpression(PyLegendBinaryExpression, PyLegendExpressionBooleanReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return ComparisonExpression(expression1, expression2, ComparisonOperator.GREATER_THAN_OR_EQUAL)

    @staticmethod
    def __to_pure_func(op1_expr: str, op2_expr: str, config: FrameToPureConfig) -> str:
        return f"({op1_expr} >= {op2_expr})"

    def __init__(self, operand1: PyLegendExpressionBooleanReturn, operand2: PyLegendExpressionBooleanReturn) -> None:
        PyLegendExpressionBooleanReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendBooleanGreaterThanEqualExpression.__to_sql_func,
            PyLegendBooleanGreaterThanEqualExpression.__to_pure_func,
            non_nullable=True
        )


class PyLegendBooleanXorExpression(PyLegendBinaryExpression, PyLegendExpressionBooleanReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return ComparisonExpression(expression1, expression2, ComparisonOperator.NOT_EQUAL)

    @staticmethod
    def __to_pure_func(op1_expr: str, op2_expr: str, config: FrameToPureConfig) -> str:
        return generate_pure_functional_call("xor", [op1_expr, op2_expr])

    def __init__(self, operand1: PyLegendExpressionBooleanReturn, operand2: PyLegendExpressionBooleanReturn) -> None:
        PyLegendExpressionBooleanReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendBooleanXorExpression.__to_sql_func,
            PyLegendBooleanXorExpression.__to_pure_func,
            non_nullable=True,
            first_operand_needs_to_be_non_nullable=True,
            second_operand_needs_to_be_non_nullable=True
        )


class PyLegendBooleanNotExpression(PyLegendUnaryExpression, PyLegendExpressionBooleanReturn):

    @staticmethod
    def __to_sql_func(
            expression: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return NotExpression(expression)

    @staticmethod
    def __to_pure_func(op_expr: str, config: FrameToPureConfig) -> str:
        return generate_pure_functional_call("not", [op_expr])

    def __init__(self, operand: PyLegendExpressionBooleanReturn) -> None:
        PyLegendExpressionBooleanReturn.__init__(self)
        PyLegendUnaryExpression.__init__(
            self,
            operand,
            PyLegendBooleanNotExpression.__to_sql_func,
            PyLegendBooleanNotExpression.__to_pure_func,
            non_nullable=True,
            operand_needs_to_be_non_nullable=True,
        )



# Copyright 2026 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import PyLegendSequence
from pylegend.core.tds.legendql_api.frames.legendql_api_tds_frame import LegendQLApiTdsFrame
from pylegend.extensions.tds.legendql_api.frames.legendql_api_legend_service_input_frame import (
    LegendQLApiLegendServiceInputFrame
)
from pylegend.samples.local_legend_env import get_local_legend_env, NORTHWIND_PROJECT_COORDINATES


__all__: PyLegendSequence[str] = [
    "northwind_orders_frame",
]


def northwind_orders_frame() -> LegendQLApiTdsFrame:
    local_legend_env = get_local_legend_env()
    return LegendQLApiLegendServiceInputFrame(
        pattern="/allOrders",
        project_coordinates=NORTHWIND_PROJECT_COORDINATES,
        legend_client=local_legend_env.legend_client,
    )



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from abc import ABCMeta, abstractmethod
from datetime import date, datetime
from decimal import Decimal as PythonDecimal
from typing import TYPE_CHECKING, overload

from typing_extensions import Concatenate

try:
    from typing import ParamSpec
except Exception:
    from typing_extensions import ParamSpec  # type: ignore

import pandas as pd

from pylegend._typing import (
    PyLegendSequence,
    PyLegendTypeVar,
    PyLegendType,
    PyLegendList,
    PyLegendTuple,
    PyLegendSet,
    PyLegendOptional,
    PyLegendCallable,
    PyLegendUnion,
    PyLegendDict
)
from pylegend.core.database.sql_to_string import (
    SqlToStringConfig,
    SqlToStringFormat
)
from pylegend.core.language import (
    PyLegendPrimitive,
    PyLegendInteger,
    PyLegendBoolean,
)
from pylegend.core.language.pandas_api.pandas_api_aggregate_specification import PyLegendAggInput
from pylegend.core.language.pandas_api.pandas_api_tds_row import PandasApiTdsRow
from pylegend.core.language.shared.primitives.primitive import PyLegendPrimitiveOrPythonPrimitive
from pylegend.core.language.shared.tds_row import AbstractTdsRow
from pylegend.core.sql.metamodel import QuerySpecification
from pylegend.core.tds.abstract.frames.base_tds_frame import BaseTdsFrame
from pylegend.core.tds.pandas_api.frames.pandas_api_tds_frame import PandasApiTdsFrame
from pylegend.core.tds.result_handler import (
    ResultHandler,
    ToStringResultHandler,
)
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.tds.tds_frame import FrameToPureConfig
from pylegend.core.tds.tds_frame import FrameToSqlConfig, PyLegendTdsFrame
from pylegend.extensions.tds.result_handler import (
    ToPandasDfResultHandler,
    PandasDfReadConfig,
)

if TYPE_CHECKING:
    from pylegend.core.language.pandas_api.pandas_api_series import Series
    from pylegend.core.tds.pandas_api.frames.pandas_api_groupby_tds_frame import PandasApiGroupbyTdsFrame
    from pylegend.core.tds.pandas_api.frames.functions.iloc import PandasApiIlocIndexer
    from pylegend.core.tds.pandas_api.frames.functions.loc import PandasApiLocIndexer
    from pylegend.core.tds.cast_helpers import CastTarget

__all__: PyLegendSequence[str] = [
    "PandasApiBaseTdsFrame"
]

R = PyLegendTypeVar('R')
P = ParamSpec("P")


class PandasApiBaseTdsFrame(PandasApiTdsFrame, BaseTdsFrame, metaclass=ABCMeta):
    __columns: PyLegendSequence[TdsColumn]

    def __init__(self, columns: PyLegendSequence[TdsColumn]) -> None:
        col_names = [c.get_name() for c in columns]
        if len(col_names) != len(set(col_names)):
            cols = "[" + ", ".join([str(c) for c in columns]) + "]"
            raise ValueError(f"TdsFrame cannot have duplicated column names. Passed columns: {cols}")
        self.__columns = [c.copy() for c in columns]
        self._transformed_frame = None

    def columns(self) -> PyLegendSequence[TdsColumn]:
        if self._transformed_frame is None:
            return [c.copy() for c in self.__columns]
        return self._transformed_frame.columns()

    @overload  # type: ignore[override]
    def __getitem__(self, key: str) -> "Series":
        ...

    @overload
    def __getitem__(self, key: PyLegendList[str]) -> "PandasApiTdsFrame":
        ...

    def __getitem__(
            self,
            key: PyLegendUnion[str, PyLegendList[str], PyLegendBoolean]
    ) -> PyLegendUnion["PandasApiTdsFrame", "Series"]:
        from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import \
            PandasApiAppliedFunctionTdsFrame
        from pylegend.core.tds.pandas_api.frames.functions.filtering import \
            PandasApiFilteringFunction
        from pylegend.core.language.shared.primitives.boolean import PyLegendBoolean

        if isinstance(key, PyLegendBoolean):
            return PandasApiAppliedFunctionTdsFrame(
                PandasApiFilteringFunction(self, filter_expr=key)
            )

        elif isinstance(key, str):
            for col in self.__columns:
                if col.get_name() == key:
                    col_type = col.get_type()
                    if col_type == "Boolean":
                        from pylegend.core.language.pandas_api.pandas_api_series import \
                            BooleanSeries  # pragma: no cover
                        return BooleanSeries(self, key)  # pragma: no cover (Boolean column not supported in PURE)
                    elif col_type in ("String", "Varchar"):
                        from pylegend.core.language.pandas_api.pandas_api_series import StringSeries
                        return StringSeries(self, key)
                    elif col_type == "Number":  # pragma: no cover
                        from pylegend.core.language.pandas_api.pandas_api_series import NumberSeries
                        return NumberSeries(self, key)
                    elif col_type in ("Integer", "TinyInt", "UTinyInt", "SmallInt", "USmallInt",
                                      "Int", "UInt", "BigInt", "UBigInt"):
                        from pylegend.core.language.pandas_api.pandas_api_series import IntegerSeries
                        return IntegerSeries(self, key)
                    elif col_type in ("Float", "Float4", "Double"):
                        from pylegend.core.language.pandas_api.pandas_api_series import FloatSeries
                        return FloatSeries(self, key)
                    elif col_type in ("Decimal", "Numeric"):
                        from pylegend.core.language.pandas_api.pandas_api_series import DecimalSeries
                        return DecimalSeries(self, key)
                    elif col_type == "Date":
                        from pylegend.core.language.pandas_api.pandas_api_series import DateSeries
                        return DateSeries(self, key)
                    elif col_type in ("DateTime", "Timestamp"):
                        from pylegend.core.language.pandas_api.pandas_api_series import DateTimeSeries
                        return DateTimeSeries(self, key)
                    elif col_type == "StrictDate":
                        from pylegend.core.language.pandas_api.pandas_api_series import StrictDateSeries
                        return StrictDateSeries(self, key)
                    else:
                        raise ValueError(f"Unsupported column type '{col_type}' for column '{key}'")  # pragma: no cover
            raise KeyError(f"['{key}'] not in index")

        elif isinstance(key, list):
            valid_col_names = {col.get_name() for col in self.__columns}
            invalid_cols = [k for k in key if k not in valid_col_names]
            if invalid_cols:
                raise KeyError(f"{invalid_cols} not in index")
            return self.filter(items=key)
        else:
            raise TypeError(f"Invalid key type: {type(key)}. Expected str, list, or boolean expression")

    def __setitem__(self, key: str, value: PyLegendUnion["Series", PyLegendPrimitiveOrPythonPrimitive]) -> None:
        """
        Pandas-like column assignment with replace semantics:
        - If column exists, drop it first.
        - Then assign the new value (Series or constant).
        """
        from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import (
            PandasApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.pandas_api.frames.functions.assign_function import AssignFunction
        from pylegend.core.language.pandas_api.pandas_api_series import Series

        # Type Check
        if not isinstance(key, str):
            raise TypeError(f"Column name must be a string, got: {type(key)}")

        # Reject cross-frame assignment
        if isinstance(value, Series):
            origin = value.get_base_frame()
            if origin is not None and origin is not self:
                raise ValueError("Assignment from a different frame is not allowed")

        # Normalize the assignment value
        col_def = {}
        if callable(value):
            col_def[key] = value
        else:
            col_def[key] = lambda row: value

        working_frame = copy.deepcopy(self)
        assign_applied = PandasApiAppliedFunctionTdsFrame(AssignFunction(working_frame, col_definitions=col_def))

        self._transformed_frame = assign_applied  # type: ignore
        self.__columns = assign_applied.columns()

    def cast(
            self,
            column_type_map: PyLegendDict[str, "CastTarget"]
    ) -> "PandasApiTdsFrame":
        from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import (
            PandasApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.pandas_api.frames.functions.cast_function import (
            PandasApiCastFunction
        )
        return PandasApiAppliedFunctionTdsFrame(PandasApiCastFunction(self, column_type_map))

    def assign(
            self,
            **kwargs: PyLegendCallable[
                [PandasApiTdsRow],
                PyLegendUnion[int, float, bool, str, date, datetime, PythonDecimal, PyLegendPrimitive]
            ],
    ) -> "PandasApiTdsFrame":
        from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import (
            PandasApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.pandas_api.frames.functions.assign_function import AssignFunction
        return PandasApiAppliedFunctionTdsFrame(AssignFunction(self, col_definitions=kwargs))

    def filter(
            self,
            items: PyLegendOptional[PyLegendList[str]] = None,
            like: PyLegendOptional[str] = None,
            regex: PyLegendOptional[str] = None,
            axis: PyLegendOptional[PyLegendUnion[str, int, PyLegendInteger]] = None
    ) -> "PandasApiTdsFrame":
        from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import (
            PandasApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.pandas_api.frames.functions.filter import PandasApiFilterFunction
        return PandasApiAppliedFunctionTdsFrame(
            PandasApiFilterFunction(
                self,
                items=items,
                like=like,
                regex=regex,
                axis=axis
            )
        )

    def sort_values(
            self,
            by: PyLegendUnion[str, PyLegendList[str]],
            axis: PyLegendUnion[str, int] = 0,
            ascending: PyLegendUnion[bool, PyLegendList[bool]] = True,
            inplace: bool = False,
            kind: PyLegendOptional[str] = None,
            na_position: str = 'last',
            ignore_index: bool = True,
            key: PyLegendOptional[PyLegendCallable[[AbstractTdsRow], AbstractTdsRow]] = None
    ) -> "PandasApiTdsFrame":
        from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import (
            PandasApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.pandas_api.frames.functions.sort_values_function import SortValuesFunction
        return PandasApiAppliedFunctionTdsFrame(SortValuesFunction(
            base_frame=self,
            by=by,
            axis=axis,
            ascending=ascending,
            inplace=inplace,
            kind=kind,
            na_position=na_position,
            ignore_index=ignore_index,
            key=key
        ))

    def truncate(
            self,
            before: PyLegendUnion[date, str, int, None] = None,
            after: PyLegendUnion[date, str, int, None] = None,
            axis: PyLegendUnion[str, int] = 0,
            copy: bool = True
    ) -> "PandasApiTdsFrame":
        from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import (
            PandasApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.pandas_api.frames.functions.truncate_function import TruncateFunction
        return PandasApiAppliedFunctionTdsFrame(TruncateFunction(
            base_frame=self,
            before=before,
            after=after,
            axis=axis,
            copy=copy
        ))

    def drop(
            self,
            labels: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str], PyLegendSet[str]]] = None,
            axis: PyLegendUnion[str, int, PyLegendInteger] = 1,
            index: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str], PyLegendSet[str]]] = None,
            columns: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str], PyLegendSet[str]]] = None,
            level: PyLegendOptional[PyLegendUnion[int, PyLegendInteger, str]] = None,
            inplace: PyLegendUnion[bool, PyLegendBoolean] = False,
            errors: str = "raise",
    ) -> "PandasApiTdsFrame":
        from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import \
            PandasApiAppliedFunctionTdsFrame
        from pylegend.core.tds.pandas_api.frames.functions.drop import PandasApiDropFunction

        return PandasApiAppliedFunctionTdsFrame(
            PandasApiDropFunction(
                base_frame=self,
                labels=labels,
                axis=axis,
                index=index,
                columns=columns,
                level=level,
                inplace=inplace,
                errors=errors
            )
        )

    def aggregate(
            self,
            func: PyLegendAggInput,
            axis: PyLegendUnion[int, str] = 0,
            *args: PyLegendPrimitiveOrPythonPrimitive,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import (
            PandasApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.pandas_api.frames.functions.aggregate_function import AggregateFunction
        return PandasApiAppliedFunctionTdsFrame(AggregateFunction(
            self,
            func,
            axis,
            *args,
            **kwargs
        ))

    def agg(
            self,
            func: PyLegendAggInput,
            axis: PyLegendUnion[int, str] = 0,
            *args: PyLegendPrimitiveOrPythonPrimitive,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import (
            PandasApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.pandas_api.frames.functions.aggregate_function import AggregateFunction
        return PandasApiAppliedFunctionTdsFrame(AggregateFunction(
            self,
            func,
            axis,
            *args,
            **kwargs
        ))

    def sum(
            self,
            axis: PyLegendUnion[int, str] = 0,
            skipna: bool = True,
            numeric_only: bool = False,
            min_count: int = 0,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        if axis not in [0, "index"]:
            raise NotImplementedError(f"The 'axis' parameter must be 0 or 'index' in sum function, but got: {axis}")
        if skipna is not True:
            raise NotImplementedError("skipna=False is not currently supported in sum function. "
                                      "SQL aggregation ignores nulls by default.")
        if numeric_only is not False:
            raise NotImplementedError("numeric_only=True is not currently supported in sum function.")
        if min_count != 0:
            raise NotImplementedError(f"min_count must be 0 in sum function, but got: {min_count}")
        if len(kwargs) > 0:
            raise NotImplementedError(
                f"Additional keyword arguments not supported in sum function: {list(kwargs.keys())}")
        return self.aggregate("sum", 0)

    def mean(
            self,
            axis: PyLegendUnion[int, str] = 0,
            skipna: bool = True,
            numeric_only: bool = False,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        if axis not in [0, "index"]:
            raise NotImplementedError(f"The 'axis' parameter must be 0 or 'index' in mean function, but got: {axis}")
        if skipna is not True:
            raise NotImplementedError("skipna=False is not currently supported in mean function.")
        if numeric_only is not False:
            raise NotImplementedError("numeric_only=True is not currently supported in mean function.")
        if len(kwargs) > 0:
            raise NotImplementedError(
                f"Additional keyword arguments not supported in mean function: {list(kwargs.keys())}")
        return self.aggregate("mean", 0)

    def min(
            self,
            axis: PyLegendUnion[int, str] = 0,
            skipna: bool = True,
            numeric_only: bool = False,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        if axis not in [0, "index"]:
            raise NotImplementedError(f"The 'axis' parameter must be 0 or 'index' in min function, but got: {axis}")
        if skipna is not True:
            raise NotImplementedError("skipna=False is not currently supported in min function.")
        if numeric_only is not False:
            raise NotImplementedError("numeric_only=True is not currently supported in min function.")
        if len(kwargs) > 0:
            raise NotImplementedError(
                f"Additional keyword arguments not supported in min function: {list(kwargs.keys())}")
        return self.aggregate("min", 0)

    def max(
            self,
            axis: PyLegendUnion[int, str] = 0,
            skipna: bool = True,
            numeric_only: bool = False,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        if axis not in [0, "index"]:
            raise NotImplementedError(f"The 'axis' parameter must be 0 or 'index' in max function, but got: {axis}")
        if skipna is not True:
            raise NotImplementedError("skipna=False is not currently supported in max function.")
        if numeric_only is not False:
            raise NotImplementedError("numeric_only=True is not currently supported in max function.")
        if len(kwargs) > 0:
            raise NotImplementedError(
                f"Additional keyword arguments not supported in max function: {list(kwargs.keys())}")
        return self.aggregate("max", 0)

    def std(
            self,
            axis: PyLegendUnion[int, str] = 0,
            skipna: bool = True,
            ddof: int = 1,
            numeric_only: bool = False,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        if axis not in [0, "index"]:
            raise NotImplementedError(f"The 'axis' parameter must be 0 or 'index' in std function, but got: {axis}")
        if skipna is not True:
            raise NotImplementedError("skipna=False is not currently supported in std function.")
        if ddof != 1:
            raise NotImplementedError(
                f"Only ddof=1 (Sample Standard Deviation) is supported in std function, but got: {ddof}")
        if numeric_only is not False:
            raise NotImplementedError("numeric_only=True is not currently supported in std function.")
        if len(kwargs) > 0:
            raise NotImplementedError(
                f"Additional keyword arguments not supported in std function: {list(kwargs.keys())}")
        return self.aggregate("std", 0)

    def var(
            self,
            axis: PyLegendUnion[int, str] = 0,
            skipna: bool = True,
            ddof: int = 1,
            numeric_only: bool = False,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        if axis not in [0, "index"]:
            raise NotImplementedError(f"The 'axis' parameter must be 0 or 'index' in var function, but got: {axis}")
        if skipna is not True:
            raise NotImplementedError("skipna=False is not currently supported in var function.")
        if ddof != 1:
            raise NotImplementedError(f"Only ddof=1 (Sample Variance) is supported in var function, but got: {ddof}")
        if numeric_only is not False:
            raise NotImplementedError("numeric_only=True is not currently supported in var function.")
        if len(kwargs) > 0:
            raise NotImplementedError(
                f"Additional keyword arguments not supported in var function: {list(kwargs.keys())}")
        return self.aggregate("var", 0)

    def count(
            self,
            axis: PyLegendUnion[int, str] = 0,
            numeric_only: bool = False,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        if axis not in [0, "index"]:
            raise NotImplementedError(f"The 'axis' parameter must be 0 or 'index' in count function, but got: {axis}")
        if numeric_only is not False:
            raise NotImplementedError("numeric_only=True is not currently supported in count function.")
        if len(kwargs) > 0:
            raise NotImplementedError(
                f"Additional keyword arguments not supported in count function: {list(kwargs.keys())}")
        return self.aggregate("count", 0)

    def groupby(
            self,
            by: PyLegendUnion[str, PyLegendList[str]],
            level: PyLegendOptional[PyLegendUnion[str, int, PyLegendList[str]]] = None,
            as_index: bool = False,
            sort: bool = True,
            group_keys: bool = False,
            observed: bool = False,
            dropna: bool = False,
    ) -> "PandasApiGroupbyTdsFrame":
        from pylegend.core.tds.pandas_api.frames.pandas_api_groupby_tds_frame import (
            PandasApiGroupbyTdsFrame
        )
        return PandasApiGroupbyTdsFrame(
            base_frame=self,
            by=by,
            level=level,
            as_index=as_index,
            sort=sort,
            group_keys=group_keys,
            observed=observed,
            dropna=dropna
        )

    def merge(
            self,
            other: "PandasApiTdsFrame",
            how: PyLegendOptional[str] = "inner",
            on: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str]]] = None,
            left_on: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str]]] = None,
            right_on: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str]]] = None,
            left_index: PyLegendOptional[bool] = False,
            right_index: PyLegendOptional[bool] = False,
            sort: PyLegendOptional[bool] = False,
            suffixes: PyLegendOptional[
                PyLegendUnion[
                    PyLegendTuple[PyLegendUnion[str, None], PyLegendUnion[str, None]],
                    PyLegendList[PyLegendUnion[str, None]],
                ]
            ] = ("_x", "_y"),
            indicator: PyLegendOptional[PyLegendUnion[bool, str]] = False,
            validate: PyLegendOptional[str] = None
    ) -> "PandasApiTdsFrame":
        """
        Pandas-like merge:
        - Mutually exclusive: `on` vs (`left_on`, `right_on`)
        - If no keys provided, infer intersection of column names
        - `how`: inner | left | right | outer (outer mapped to full)
        - `suffixes`: applied to overlapping non-key columns
        """
        from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import (
            PandasApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.pandas_api.frames.functions.merge import (
            PandasApiMergeFunction
        )
        merge_fn = PandasApiMergeFunction(
            self,
            other,  # type: ignore
            how=how,
            on=on,
            left_on=left_on,
            right_on=right_on,
            left_index=left_index,
            right_index=right_index,
            sort=sort,
            suffixes=suffixes,
            indicator=indicator,
            validate=validate
        )
        merged = PandasApiAppliedFunctionTdsFrame(merge_fn)

        if sort:
            return merged.sort_values(
                by=merge_fn.get_sort_keys(),
                axis=0,
                ascending=True,
                inplace=False,
                kind=None,
                na_position="last",
                ignore_index=True,
                key=None
            )
        else:
            return merged

    def join(
            self,
            other: "PandasApiTdsFrame",
            on: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str]]] = None,
            how: PyLegendOptional[str] = "left",
            lsuffix: str = "",
            rsuffix: str = "",
            sort: PyLegendOptional[bool] = False,
            validate: PyLegendOptional[str] = None
    ) -> "PandasApiTdsFrame":
        """
        Pandas-like join delegating to merge. No index support, only column-on-column via `on`.
        """
        return self.merge(
            other=other,
            how=how,
            on=on,
            sort=sort,
            suffixes=[lsuffix, rsuffix],
            validate=validate
        )

    def rename(
            self,
            mapper: PyLegendOptional[PyLegendUnion[PyLegendDict[str, str], PyLegendCallable[[str], str]]] = None,
            index: PyLegendOptional[PyLegendUnion[PyLegendDict[str, str], PyLegendCallable[[str], str]]] = None,
            columns: PyLegendOptional[PyLegendUnion[PyLegendDict[str, str], PyLegendCallable[[str], str]]] = None,
            axis: PyLegendUnion[str, int] = 1,
            inplace: PyLegendUnion[bool] = False,
            copy: PyLegendUnion[bool] = True,
            level: PyLegendOptional[PyLegendUnion[int, str]] = None,
            errors: str = "ignore",
    ) -> "PandasApiTdsFrame":
        """
        Pandas-like rename:
        - Supports mapping via `mapper` or explicit `index`/`columns`
        - Only column renames are applied when `axis` is 1
        - `errors`: ignore | raise
        """

        from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import (
            PandasApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.pandas_api.frames.functions.rename import (
            PandasApiRenameFunction
        )
        return PandasApiAppliedFunctionTdsFrame(
            PandasApiRenameFunction(
                base_frame=self,
                mapper=mapper,
                axis=axis,
                index=index,
                columns=columns,
                copy=copy,
                inplace=inplace,
                level=level,
                errors=errors
            )
        )

    def apply(
            self,
            func: PyLegendUnion[
                PyLegendCallable[Concatenate["Series", P], PyLegendPrimitiveOrPythonPrimitive],
                str
            ],
            axis: PyLegendUnion[int, str] = 0,
            raw: bool = False,
            result_type: PyLegendOptional[str] = None,
            args: PyLegendTuple[PyLegendPrimitiveOrPythonPrimitive, ...] = (),
            by_row: PyLegendUnion[bool, str] = "compat",
            engine: str = "python",
            engine_kwargs: PyLegendOptional[PyLegendDict[str, PyLegendPrimitiveOrPythonPrimitive]] = None,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        """
        Pandas-like apply (columns-only):
        - Supports callable func applied to each column (axis=0 or 'index')
        - Internally delegates to assign by constructing lambdas per column
        - Unsupported params raise NotImplementedError
        """

        from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import (
            PandasApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.pandas_api.frames.functions.assign_function import AssignFunction
        from pylegend.core.language.pandas_api.pandas_api_series import Series

        # Validation
        if axis not in (0, "index"):
            raise ValueError("Only column-wise apply is supported. Use axis=0 or 'index'")
        if raw:
            raise NotImplementedError("raw=True is not supported. Use raw=False")
        if result_type is not None:
            raise NotImplementedError("result_type is not supported")
        if by_row not in (False, "compat"):
            raise NotImplementedError("by_row must be False or 'compat'")
        if engine != "python":
            raise NotImplementedError("Only engine='python' is supported")
        if engine_kwargs is not None:
            raise NotImplementedError("engine_kwargs are not supported")
        if isinstance(func, str):
            raise NotImplementedError("String-based apply is not supported")
        if not callable(func):
            raise TypeError("Function must be a callable")

        # Build assign column definitions: apply func to each column Series
        col_definitions = {}
        for c in self.columns():
            col_name = c.get_name()
            series = self[col_name]

            # Compute row callable via func on the Series
            def _row_callable(
                    _row: PandasApiTdsRow,
                    _s: Series = series,
                    _a: PyLegendTuple[PyLegendPrimitiveOrPythonPrimitive, ...] = args,
                    _k: PyLegendPrimitiveOrPythonPrimitive = kwargs  # type: ignore
            ) -> PyLegendPrimitiveOrPythonPrimitive:
                return func(_s, *_a, **_k)  # type: ignore

            col_definitions[col_name] = _row_callable

        return PandasApiAppliedFunctionTdsFrame(
            AssignFunction(self, col_definitions=col_definitions)  # type: ignore
        )

    @property
    def iloc(self) -> "PandasApiIlocIndexer":
        """
        Purely integer-location based indexing for selection by position.
        .iloc[] is primarily integer position based (from 0 to length-1 of the axis).

        Allowed inputs are:
        - An integer, e.g. 5.
        - A slice object with ints, e.g. 1:7.
        - A tuple of row and column indexes, e.g., (slice(1, 5), slice(0, 2))

        Other pandas iloc features such as list of integers, boolean arrays, and callables
        are not supported and will raise a NotImplementedError.
        """
        from pylegend.core.tds.pandas_api.frames.functions.iloc import PandasApiIlocIndexer
        return PandasApiIlocIndexer(self)

    @property
    def loc(self) -> "PandasApiLocIndexer":
        """
        Access a group of rows and columns by label(s) or a boolean array.
        .loc[] is primarily label based, but may also be used with a boolean array.

        Allowed inputs are:
        - A single label, e.g. 5 or 'a', (note that 5 is interpreted as a
          label of the index, not as an integer position along the index).
        - A list or array of labels, e.g. ['a', 'b', 'c'].
        - A slice object with labels, e.g. 'a':'f'.
        - A boolean array of the same length as the axis being sliced.
        - A callable function with one argument (the calling Series or
          DataFrame) and that returns valid output for indexing (one of the above).

        Currently, for row selection, only callable function or complete slice are supported.
        For column selection, string labels, lists of string labels, and slices of string labels are supported.
        """
        from pylegend.core.tds.pandas_api.frames.functions.loc import PandasApiLocIndexer
        return PandasApiLocIndexer(self)

    def head(self, n: int = 5) -> "PandasApiTdsFrame":
        """
        Return the first `n` rows by calling truncate on rows.
        Negative `n` is not supported.
        """
        if not isinstance(n, int):
            raise TypeError(f"n must be an int, got {type(n)}")
        if n < 0:
            raise NotImplementedError("Negative n is not supported yet in Pandas API head")

        return self.truncate(before=None, after=max(n - 1, -1), axis=0, copy=True)

    @property
    def shape(self) -> PyLegendTuple[int, int]:
        """
        Return a tuple representing the dimensionality of the TdsFrame
        as (number of rows, number of columns).
        """

        col_name = self.columns()[0].get_name()
        newframe = self.aggregate(func={col_name: "count"}, axis=0)

        df = newframe.execute_frame_to_pandas_df()

        total_rows = df.iloc[0, 0]
        total_cols = len(self.columns())

        return (total_rows, total_cols)  # type: ignore

    def dropna(
            self,
            axis: PyLegendUnion[int, str] = 0,
            how: str = "any",
            thresh: PyLegendOptional[int] = None,
            subset: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str]]] = None,
            inplace: bool = False,
            ignore_index: bool = False
    ) -> "PandasApiTdsFrame":
        """
        Remove missing values.

        Parameters
        ----------
        axis : {0 or 'index'}, default 0
            Determine if rows or columns which contain missing values are removed.
            * 0, or 'index' : Drop rows which contain missing values.
            Currently, only `axis=0` is supported.
        how : {'any', 'all'}, default 'any'
            Determine if row is removed from TdsFrame, when we have at least one NA or all NA.
            * 'any' : If any NA values are present, drop that row.
            * 'all' : If all values are NA, drop that row.
        thresh : int, optional
            Not implemented yet.
        subset : list-like, optional
            Labels along other axis to consider, e.g. if you are dropping rows
            these would be a list of columns to include.
        inplace : bool, default False
            Not implemented yet.
        ignore_index : bool, default False
            Not implemented yet.

        Returns
        -------
        PandasApiTdsFrame
            TdsFrame with NA entries dropped.
        """
        from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import (
            PandasApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.pandas_api.frames.functions.dropna import PandasApiDropnaFunction
        return PandasApiAppliedFunctionTdsFrame(
            PandasApiDropnaFunction(
                base_frame=self,
                axis=axis,
                how=how,
                thresh=thresh,
                subset=subset,
                inplace=inplace,
                ignore_index=ignore_index
            )
        )

    def fillna(
            self,
            value: PyLegendUnion[
                int, float, str, bool, date, datetime,
                PyLegendDict[str, PyLegendUnion[int, float, str, bool, date, datetime]]
            ] = None,  # type: ignore
            axis: PyLegendOptional[PyLegendUnion[int, str]] = 0,
            inplace: bool = False,
            limit: PyLegendOptional[int] = None
    ) -> "PandasApiTdsFrame":
        """
        Fill missing values.

        Parameters
        ----------
        base_frame : PandasApiBaseTdsFrame
            The base frame to apply fillna on.
        value : scalar, dict, default None
            Value to use to fill holes (e.g. 0), alternately a dict of values specifying
            which value to use for each column of TdsFrame.
        axis : {0 or 'index'}, default 0
            Axis along which to fill missing values.
            * 0, or 'index' : Fill missing values for each column.
            Currently, only `axis=0` is supported.
        inplace : bool, default False
            Not implemented yet.
        limit : int, optional
            Not implemented yet.

        Returns
        -------
        PandasApiTdsFrame
            TdsFrame with NA entries filled.
        """
        from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import (
            PandasApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.pandas_api.frames.functions.fillna import PandasApiFillnaFunction
        return PandasApiAppliedFunctionTdsFrame(
            PandasApiFillnaFunction(
                base_frame=self,
                value=value,
                axis=axis,
                inplace=inplace,
                limit=limit
            )
        )

    def rank(
            self,
            axis: PyLegendUnion[int, str] = 0,
            method: str = 'min',
            numeric_only: bool = False,
            na_option: str = 'bottom',
            ascending: bool = True,
            pct: bool = False
    ) -> "PandasApiTdsFrame":
        from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import (
            PandasApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.pandas_api.frames.functions.rank_function import RankFunction
        return PandasApiAppliedFunctionTdsFrame(RankFunction(
            base_frame=self,
            axis=axis,
            method=method,
            numeric_only=numeric_only,
            na_option=na_option,
            ascending=ascending,
            pct=pct
        ))

    @abstractmethod
    def get_super_type(self) -> PyLegendType[PyLegendTdsFrame]:
        pass  # pragma: no cover

    def to_sql_query_object(self, config: FrameToSqlConfig) -> QuerySpecification:
        if self._transformed_frame is None:
            return self.get_super_type().to_sql_query_object(self, config)  # type: ignore
        else:
            return self._transformed_frame.to_sql_query_object(config)

    def to_pure(self, config: FrameToPureConfig) -> str:
        if self._transformed_frame is None:
            return self.get_super_type().to_pure(self, config)  # type: ignore
        else:
            return self._transformed_frame.to_pure(config)

    def to_pure_query(self, config: FrameToPureConfig = FrameToPureConfig()) -> str:
        return self.to_pure(config)

    @abstractmethod
    def get_all_tds_frames(self) -> PyLegendList["PandasApiBaseTdsFrame"]:
        pass  # pragma: no cover

    def to_sql_query(self, config: FrameToSqlConfig = FrameToSqlConfig()) -> str:
        query = self.to_sql_query_object(config)
        sql_to_string_config = SqlToStringConfig(
            format_=SqlToStringFormat(
                pretty=config.pretty
            )
        )
        return config.sql_to_string_generator().generate_sql_string(query, sql_to_string_config)

    def execute_frame(
            self,
            result_handler: ResultHandler[R],
            chunk_size: PyLegendOptional[int] = None
    ) -> R:
        from pylegend.core.tds.pandas_api.frames.pandas_api_input_tds_frame import (
            PandasApiInputTdsFrame,
            PandasApiExecutableInputTdsFrame
        )
        tds_frames = self.get_all_tds_frames()
        input_frames = [x for x in tds_frames if isinstance(x, PandasApiInputTdsFrame)]

        non_exec_frames = [x for x in input_frames if not isinstance(x, PandasApiExecutableInputTdsFrame)]
        if non_exec_frames:
            raise ValueError(
                "Cannot execute frame as its built on top of non-executable input frames: [" +
                (", ".join([str(f) for f in non_exec_frames]) + "]")
            )

        exec_frames = [x for x in input_frames if isinstance(x, PandasApiExecutableInputTdsFrame)]

        all_legend_clients = []
        for e in exec_frames:
            c = e.get_legend_client()
            if c not in all_legend_clients:
                all_legend_clients.append(c)
        if len(all_legend_clients) > 1:
            raise ValueError(
                "Found tds frames with multiple legend_clients (which is not supported): [" +
                (", ".join([str(f) for f in all_legend_clients]) + "]")
            )
        legend_client = all_legend_clients[0]
        result = legend_client.execute_sql_string(self.to_sql_query(), chunk_size=chunk_size)
        return result_handler.handle_result(self, result)

    def execute_frame_to_string(
            self,
            chunk_size: PyLegendOptional[int] = None
    ) -> str:
        return self.execute_frame(ToStringResultHandler(), chunk_size)

    def execute_frame_to_pandas_df(
            self,
            chunk_size: PyLegendOptional[int] = None,
            pandas_df_read_config: PandasDfReadConfig = PandasDfReadConfig()
    ) -> pd.DataFrame:
        return self.execute_frame(
            ToPandasDfResultHandler(pandas_df_read_config),
            chunk_size
        )
